# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""All built-in types of waveforms and their Waveform parent class."""

from __future__ import annotations

from abc import ABC, abstractmethod
import functools
import inspect
import itertools
import sys
from sys import version_info
from types import FunctionType
from typing import Any, cast, Optional, Tuple, Union
import warnings

from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike
import scipy.interpolate as interpolate

from pulser.parametrized import Parametrized, ParamObj
from pulser.parametrized.decorators import parametrize
from pulser.json.utils import obj_to_dict

if version_info[:2] >= (3, 8):  # pragma: no cover
    from functools import cached_property
else:  # pragma: no cover
    try:
        from backports.cached_property import cached_property  # type: ignore
    except ImportError:
        raise ImportError(
            "Using pulser with Python version 3.7 requires the"
            " `backports.cached-property` module. Install it by running"
            " `pip install backports.cached-property`."
        )


class Waveform(ABC):
    """The abstract class for a pulse's waveform."""

    def __new__(cls, *args, **kwargs):  # type: ignore
        """Creates a Waveform instance or a ParamObj depending on the input."""
        for x in itertools.chain(args, kwargs.values()):
            if isinstance(x, Parametrized):
                return ParamObj(cls, *args, **kwargs)
        else:
            return object.__new__(cls)

    def __init__(self, duration: Union[int, Parametrized]):
        """Initializes a waveform with a given duration.

        Args:
            duration (int): The waveforms duration (in ns).
        """
        duration = cast(int, duration)
        try:
            _duration = int(duration)
        except (TypeError, ValueError):
            raise TypeError(
                "duration needs to be castable to an int but "
                f"type {type(duration)} was provided."
            )
        if _duration <= 0:
            raise ValueError(
                "A waveform must have a positive duration, "
                + f"not {duration}."
            )
        elif duration - _duration != 0:
            warnings.warn(
                f"A waveform duration of {duration} ns is below the"
                " supported precision of 1 ns. It was rounded down "
                + f"to {_duration} ns.",
                stacklevel=3,
            )

        self._duration = _duration

    @property
    @abstractmethod
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        pass

    @cached_property
    @abstractmethod
    def _samples(self) -> np.ndarray:
        pass

    @property
    def samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            np.ndarray: A numpy array with a value for each time step.
        """
        return self._samples.copy()

    @property
    def first_value(self) -> float:
        """The first value in the waveform."""
        return float(self[0])

    @property
    def last_value(self) -> float:
        """The last value in the waveform."""
        return float(self[-1])

    @property
    def integral(self) -> float:
        """Integral of the waveform (time in ns, value in rad/µs)."""
        return float(np.sum(self.samples)) * 1e-3  # ns * rad/µs = 1e-3

    def draw(self) -> None:
        """Draws the waveform."""
        fig, ax = plt.subplots()
        self._plot(ax, "rad/µs")

        plt.show()

    def change_duration(self, new_duration: int) -> Waveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support"
            " modifications to its duration."
        )

    @abstractmethod
    def _to_dict(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    def __getitem__(
        self, index_or_slice: Union[int, slice]
    ) -> Union[float, np.ndarray]:
        if isinstance(index_or_slice, slice):
            s: slice = self._check_slice(index_or_slice)
            return cast(np.ndarray, self._samples[s])
        else:
            index: int = self._check_index(index_or_slice)
            return cast(float, self._samples[index])

    def _check_index(self, i: int) -> int:
        if i < -self.duration or i >= self.duration:
            raise IndexError(
                "Index ('index_or_slice' = "
                f"{i}) must be in the range "
                f"0~{self.duration-1}, or "
                f"{-self.duration}~-1 from the end."
            )
        return i if i >= 0 else self.duration + i

    def _check_slice(self, s: slice) -> slice:
        if s.step is not None and s.step != 1:
            raise IndexError("The step of the slice must be None or 1.")

        # Transform start and stop indexes into positive or null values
        # since they can be omitted (None) or negative (end-indexing)
        start = (
            0
            if s.start is None
            else (s.start if s.start >= 0 else self.duration + s.start)
        )
        stop = (
            self.duration
            if s.stop is None
            else (s.stop if s.stop >= 0 else self.duration + s.stop)
        )

        # Correct out of bounds ranges
        if start < 0:
            start = 0
        if stop < 0:
            stop = 0
        if start > self.duration:
            start = self.duration
        if stop > self.duration:
            stop = self.duration
        if stop < start:
            stop = start

        return slice(start, stop)

    @abstractmethod
    def __mul__(self, other: float) -> Waveform:
        pass

    def __neg__(self) -> Waveform:
        return self.__mul__(-1.0)

    def __truediv__(self, other: float) -> Waveform:
        if other == 0:
            raise ZeroDivisionError("Can't divide a waveform by zero.")
        else:
            return self.__mul__(1 / other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Waveform):
            return False
        elif self.duration != other.duration:
            return False
        else:
            return bool(np.all(np.isclose(self.samples, other.samples)))

    def __hash__(self) -> int:
        return hash(tuple(self.samples))

    def _plot(
        self, ax: Axes, ylabel: str, color: Optional[str] = None
    ) -> None:
        ax.set_xlabel("t (ns)")
        ts = np.arange(self.duration)
        if color:
            ax.set_ylabel(ylabel, color=color, fontsize=14)
            ax.plot(ts, self.samples, color=color)
            ax.tick_params(axis="y", labelcolor=color)
            ax.axhline(0, color=color, linestyle=":", linewidth=0.5)
        else:
            ax.set_ylabel(ylabel, fontsize=14)
            ax.plot(ts, self.samples)
            ax.axhline(0, color="black", linestyle=":", linewidth=0.5)


class CompositeWaveform(Waveform):
    """A waveform combining multiple smaller waveforms.

    Args:
        waveforms(Waveform): Two or more waveforms to combine.
    """

    def __init__(self, *waveforms: Union[Parametrized, Waveform]):
        """Initializes a waveform from multiple waveforms."""
        if len(waveforms) < 2:
            raise ValueError(
                "Needs at least two waveforms to form a " "CompositeWaveform."
            )
        waveforms = cast(Tuple[Waveform], waveforms)
        for wf in waveforms:
            self._validate(wf)

        self._waveforms = list(waveforms)

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        duration = 0
        for wf in self._waveforms:
            duration += wf.duration
        return duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return cast(
            np.ndarray, np.concatenate([wf.samples for wf in self._waveforms])
        )

    @property
    def waveforms(self) -> list[Waveform]:
        """The waveforms encapsulated in the composite waveform."""
        return list(self._waveforms)

    def _validate(self, waveform: Waveform) -> None:
        if not isinstance(waveform, Waveform):
            raise TypeError(
                f"{waveform!r} is not a valid waveform. "
                "Please provide a valid Waveform."
            )

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, *self._waveforms)

    def __str__(self) -> str:
        contents_list = ["{!r}"] * len(self._waveforms)
        contents = ", ".join(contents_list)
        contents = contents.format(*self._waveforms)
        return f"Composite({contents})"

    def __repr__(self) -> str:
        return f"CompositeWaveform({self.duration} ns, {self._waveforms!r})"

    def __mul__(self, other: float) -> CompositeWaveform:
        return CompositeWaveform(*(wf * other for wf in self._waveforms))


class CustomWaveform(Waveform):
    """A custom waveform.

    Args:
        samples (array_like): The modulation values at each time step
            (in rad/µs). The number of samples dictates the duration, in ns.
    """

    def __init__(self, samples: ArrayLike):
        """Initializes a custom waveform."""
        samples_arr = np.array(samples, dtype=float)
        self._samples: np.ndarray = samples_arr
        super().__init__(len(samples_arr))

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        # self._samples is already cached when initialized in __init__
        pass

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._samples)

    def __str__(self) -> str:
        return "Custom"

    def __repr__(self) -> str:
        return f"CustomWaveform({self.duration} ns, {self.samples!r})"

    def __mul__(self, other: float) -> CustomWaveform:
        return CustomWaveform(self._samples * float(other))


class ConstantWaveform(Waveform):
    """A waveform of constant value.

    Args:
        duration (int): The waveform duration (in ns).
        value (float): The modulation value (in rad/µs).
    """

    def __init__(
        self,
        duration: Union[int, Parametrized],
        value: Union[float, Parametrized],
    ):
        """Initializes a constant waveform."""
        super().__init__(duration)
        value = cast(float, value)
        self._value = float(value)

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return np.full(self.duration, self._value)

    def change_duration(self, new_duration: int) -> ConstantWaveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.

        Returns:
            ConstantWaveform: The new waveform with the given duration.
        """
        return ConstantWaveform(new_duration, self._value)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._duration, self._value)

    def __str__(self) -> str:
        return f"{self._value:.3g} rad/µs"

    def __repr__(self) -> str:
        return (
            f"ConstantWaveform({self._duration} ns, "
            + f"{self._value:.3g} rad/µs)"
        )

    def __mul__(self, other: float) -> ConstantWaveform:
        return ConstantWaveform(self._duration, self._value * float(other))


class RampWaveform(Waveform):
    """A linear ramp waveform.

    Args:
        duration (int): The waveform duration (in ns).
        start (float): The initial value (in rad/µs).
        stop (float): The final value (in rad/µs).
    """

    def __init__(
        self,
        duration: Union[int, Parametrized],
        start: Union[float, Parametrized],
        stop: Union[float, Parametrized],
    ):
        """Initializes a ramp waveform."""
        super().__init__(duration)
        start = cast(float, start)
        self._start: float = float(start)
        stop = cast(float, stop)
        self._stop: float = float(stop)

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return np.linspace(self._start, self._stop, num=self._duration)

    @property
    def slope(self) -> float:
        r"""Slope of the ramp, in :math:`s^{-15}`."""
        return (self._stop - self._start) / self._duration

    def change_duration(self, new_duration: int) -> RampWaveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.

        Returns:
            RampWaveform: The new waveform with the given duration.
        """
        return RampWaveform(new_duration, self._start, self._stop)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._duration, self._start, self._stop)

    def __str__(self) -> str:
        return f"Ramp({self._start:.3g}->{self._stop:.3g} rad/µs)"

    def __repr__(self) -> str:
        return (
            f"RampWaveform({self._duration} ns, "
            + f"{self._start:.3g}->{self._stop:.3g} rad/µs)"
        )

    def __mul__(self, other: float) -> RampWaveform:
        k = float(other)
        return RampWaveform(self._duration, self._start * k, self._stop * k)


class BlackmanWaveform(Waveform):
    """A Blackman window of a specified duration and area.

    Args:
        duration (int): The waveform duration (in ns).
        area (float): The integral of the waveform. Can be negative, in which
            case it takes the positive waveform and changes the sign of all its
            values.
    """

    def __init__(
        self,
        duration: Union[int, Parametrized],
        area: Union[float, Parametrized],
    ):
        """Initializes a Blackman waveform."""
        super().__init__(duration)
        try:
            self._area: float = float(cast(float, area))
        except (TypeError, ValueError):
            raise TypeError(
                "area needs to be castable to a float but "
                f"type {type(area)} was provided."
            )

        self._norm_samples: np.ndarray = np.clip(
            np.blackman(self._duration), 0, np.inf
        )
        self._scaling: float = (
            self._area / float(np.sum(self._norm_samples)) / 1e-3
        )

    @classmethod
    @parametrize
    def from_max_val(
        cls,
        max_val: Union[float, Parametrized],
        area: Union[float, Parametrized],
    ) -> BlackmanWaveform:
        """Creates a Blackman waveform with a threshold on the maximum value.

        Instead of defining a duration, the waveform is defined by its area and
        the maximum value. The duration is chosen so that the maximum value is
        not surpassed, but approached as closely as possible.

        Args:
            max_val (float): The maximum value threshold (in rad/µs). If
                negative, it is taken as the lower bound i.e. the minimum
                value that can be reached. The sign of `max_val` must match the
                sign of `area`.
            area (float): The area under the waveform.
        """
        max_val = cast(float, max_val)
        area = cast(float, area)
        area_sign = np.sign(area)
        if np.sign(max_val) != area_sign:
            raise ValueError(
                "The maximum value and the area must have " "matching signs."
            )

        # Deal only with positive areas
        area *= float(area_sign)
        max_val *= float(area_sign)

        # A normalized Blackman waveform has an area of 0.42 * duration
        duration = np.ceil(area / (0.42 * max_val) * 1e3)  # in ns
        wf = cls(duration, area)
        previous_wf = None

        # Adjust for rounding errors to make sure max_val is not surpassed
        while wf._scaling > max_val:
            duration += 1
            previous_wf = wf
            wf = cls(duration, area)

        # According to the documentation for numpy.blackman(), "the value one
        # appears only if the number of samples is odd". Hence, the previous
        # even duration can reach a maximal value closer to max_val.
        if (
            previous_wf is not None
            and duration % 2 == 1
            and np.max(wf.samples) < np.max(previous_wf.samples) <= max_val
        ):
            wf = previous_wf

        # Restore original sign to the waveform
        return wf if area_sign != -1 else cast(BlackmanWaveform, -wf)

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return cast(np.ndarray, self._norm_samples * self._scaling)

    def change_duration(self, new_duration: int) -> BlackmanWaveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.

        Returns:
            BlackmanWaveform: The new waveform with the same area but a new
            duration.
        """
        return BlackmanWaveform(new_duration, self._area)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._duration, self._area)

    def __str__(self) -> str:
        return f"Blackman(Area: {self._area:.3g})"

    def __repr__(self) -> str:
        return f"BlackmanWaveform({self._duration} ns, Area: {self._area:.3g})"

    def __mul__(self, other: float) -> BlackmanWaveform:
        return BlackmanWaveform(self._duration, self._area * float(other))


class InterpolatedWaveform(Waveform):
    """Creates a waveform from interpolation of a set of data points.

    Args:
        duration (int): The waveform duration (in ns).
        values (ArrayLike): Values of the interpolation points (in rad/µs).
        times (Optional[ArrayLike]): Fractions of the total duration (between 0
            and 1), indicating where to place each value on the time axis. If
            not given, the values are spread evenly throughout the full
            duration of the waveform.
        interpolator (str = "PchipInterpolator"): The SciPy interpolation class
            to use. Supports "PchipInterpolator" and "interp1d".
        **interpolator_kwargs: Extra parameters to give to the chosen
            interpolator class.
    """

    def __init__(
        self,
        duration: Union[int, Parametrized],
        values: Union[ArrayLike, Parametrized],
        times: Optional[Union[ArrayLike, Parametrized]] = None,
        interpolator: str = "PchipInterpolator",
        **interpolator_kwargs: Any,
    ):
        """Initializes a new InterpolatedWaveform."""
        super().__init__(duration)
        self._values = np.array(values, dtype=float)
        if times is not None:
            times_ = np.array(times, dtype=float)
            if len(times_) != len(self._values):
                raise ValueError(
                    "When specified, the number of time coordinates in `times`"
                    f" ({len(times_)}) must match the number of `values` "
                    f"({len(self._values)})."
                )
            if np.any(times_ < 0):
                raise ValueError(
                    "All values in `times` must be greater than or equal to 0."
                )
            if np.any(times_ > 1):
                raise ValueError(
                    "All values in `times` must be less than or equal to 1."
                )
            unique_times = np.unique(times)  # Sorted array of unique values
            if len(times_) != len(unique_times):
                raise ValueError(
                    "`times` must be an array of non-repeating values."
                )
            self._times = times_
        else:
            self._times = np.linspace(0, 1, num=len(self._values))

        valid_interpolators = ("PchipInterpolator", "interp1d")
        if interpolator not in valid_interpolators:
            raise ValueError(
                f"Invalid interpolator '{interpolator}', only "
                "accepts: " + ", ".join(valid_interpolators)
            )
        interp_cls = getattr(interpolate, interpolator)
        self._data_pts = np.array(
            [
                (round(t), v)
                for t, v in zip(
                    self._times * (self._duration - 1), self._values
                )
            ]
        )
        self._interp_func = interp_cls(
            self._data_pts[:, 0], self._data_pts[:, 1], **interpolator_kwargs
        )
        self._kwargs: dict[str, Any] = {
            "times": times,
            "interpolator": interpolator,
            **interpolator_kwargs,
        }

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform."""
        return cast(
            np.ndarray,
            np.round(
                self._interp_func(np.arange(self._duration)), decimals=9
            ),  # Rounds to the order of Hz
        )

    @property
    def interp_function(
        self,
    ) -> Union[interpolate.PchipInterpolator, interpolate.interp1d]:
        """The interpolating function."""
        return self._interp_func

    @property
    def data_points(self) -> np.ndarray:
        """Points (t[ns], value[rad/µs]) that define the interpolation."""
        return self._data_pts.copy()

    def change_duration(self, new_duration: int) -> InterpolatedWaveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.

        Returns:
            InterpolatedWaveform: The new waveform with the same coordinates
            for interpolation but a new duration.
        """
        return InterpolatedWaveform(new_duration, self._values, **self._kwargs)

    def _plot(
        self, ax: Axes, ylabel: str, color: Optional[str] = None
    ) -> None:
        super()._plot(ax, ylabel, color=color)
        ax.scatter(self._data_pts[:, 0], self._data_pts[:, 1], c=color)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._duration, self._values, **self._kwargs)

    def __str__(self) -> str:
        coords = [f"({int(x)}, {y:.4g})" for x, y in self.data_points]
        return f"InterpolatedWaveform(Points: {', '.join(coords)})"

    def __repr__(self) -> str:
        interp_str = f", Interpolator={self._kwargs['interpolator']})"
        return self.__str__()[:-1] + interp_str

    def __mul__(self, other: float) -> InterpolatedWaveform:
        return InterpolatedWaveform(
            self._duration, self._values * other, **self._kwargs
        )


class KaiserWaveform(Waveform):
    """A Kaiser window of a specified duration and beta parameter.

    For more information on the Kaiser window and the beta parameter,
    check the numpy documentation for the kaiser(M, beta) function:
    https://numpy.org/doc/stable/reference/generated/numpy.kaiser.html

    Args:
        duration (int): The waveform duration (in ns).
        area (float): The integral of the waveform. Can be negative,
            in which case it takes the positive waveform and changes the sign
            of all its values.
        beta (Optional[float]): The beta parameter of the Kaiser window.
            The default value is 14.
    """

    def __init__(
        self,
        duration: Union[int, Parametrized],
        area: Union[float, Parametrized],
        beta: Optional[Union[float, Parametrized]] = 14.0,
    ):
        """Initializes a Kaiser waveform."""
        super().__init__(duration)

        try:
            self._area: float = float(cast(float, area))
        except (TypeError, ValueError):
            raise TypeError(
                "area needs to be castable to a float but "
                f"type {type(area)} was provided."
            )

        try:
            self._beta: float = float(cast(float, beta))
        except (TypeError, ValueError):
            raise TypeError(
                "beta needs to be castable to a float but "
                f"type {type(beta)} was provided."
            )

        if self._beta < 0.0:
            raise ValueError(
                f"The beta parameter (`beta` = {self._beta})"
                " must be greater than 0."
            )

        self._norm_samples: np.ndarray = np.clip(
            np.kaiser(self._duration, self._beta), 0, np.inf
        )

        self._scaling: float = (
            self._area / float(np.sum(self._norm_samples)) / 1e-3
        )

    @classmethod
    @parametrize
    def from_max_val(
        cls,
        max_val: Union[float, Parametrized],
        area: Union[float, Parametrized],
        beta: Optional[Union[float, Parametrized]] = 14.0,
    ) -> KaiserWaveform:
        """Creates a Kaiser waveform with a threshold on the maximum value.

        Instead of defining a duration, the waveform is defined by its area and
        the maximum value. The duration is chosen so that the maximum value is
        not surpassed, but approached as closely as possible.

        Args:
            max_val (float): The maximum value threshold (in rad/µs). If
                negative, it is taken as the lower bound i.e. the minimum
                value that can be reached. The sign of `max_val` must match the
                sign of `area`.
            area (float): The area under the waveform.
            beta (Optional[float]): The beta parameter of the Kaiser window.
                The default value is 14.
        """
        max_val = cast(float, max_val)
        area = cast(float, area)

        if np.sign(max_val) != np.sign(area):
            raise ValueError(
                "The maximum value and the area must have matching signs."
            )

        # All computations will be done on a positive area

        is_negative: bool = area < 0
        if is_negative:
            area = -area
            max_val = -max_val

        # Compute the ratio area / duration for a long duration
        # and use this value for a first guess of the best duration

        ratio: float = max_val * np.sum(np.kaiser(100, beta)) / 100
        duration_guess: int = int(area * 1000.0 / ratio)

        duration_best: int = 0

        if duration_guess < 11:
            # Because of the seesawing effect on short durations,
            # all solutions must be tested to find the best one

            max_val_best: float = 0
            for duration in range(1, 16):
                kaiser_temp = np.kaiser(duration, beta)
                scaling_temp = 1000 * area / np.sum(kaiser_temp)
                max_val_temp = np.max(kaiser_temp) * scaling_temp
                if max_val_best < max_val_temp <= max_val:
                    max_val_best = max_val_temp
                    duration_best = duration

        else:

            # Start with a waveform based on the duration guess

            kaiser_guess = np.kaiser(duration_guess, beta)
            scaling_guess = 1000 * area / np.sum(kaiser_guess)
            max_val_temp = np.max(kaiser_guess) * scaling_guess

            # Increase or decrease duration depending on
            # the max value for the guessed duration

            step = 1 if np.max(kaiser_guess) * scaling_guess >= max_val else -1
            duration = duration_guess

            while np.sign(max_val_temp - max_val) == step:
                duration += step
                kaiser_temp = np.kaiser(duration, beta)
                scaling = 1000 * area / np.sum(kaiser_temp)
                max_val_temp = np.max(kaiser_temp) * scaling

            duration_best = duration if step == 1 else duration + 1

        # Restore the original area if it was negative

        if is_negative:
            area = -area

        return cls(duration_best, area, beta)

    @property
    def duration(self) -> int:
        """The duration of the pulse (in ns)."""
        return self._duration

    @cached_property
    def _samples(self) -> np.ndarray:
        """The value at each time step that describes the waveform.

        Returns:
            numpy.ndarray: A numpy array with a value for each time step.
        """
        return cast(np.ndarray, self._norm_samples * self._scaling)

    def change_duration(self, new_duration: int) -> KaiserWaveform:
        """Returns a new waveform with modified duration.

        Args:
            new_duration(int): The duration of the new waveform.

        Returns:
            KaiserWaveform: The new waveform with the same area and beta
            but a new duration.
        """
        return KaiserWaveform(new_duration, self._area, self._beta)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(self, self._duration, self._area, self._beta)

    def __str__(self) -> str:
        return (
            f"Kaiser({self._duration} ns, "
            f"Area: {self._area:.3g}, Beta: {self._beta:.3g})"
        )

    def __repr__(self) -> str:
        return (
            f"KaiserWaveform(duration: {self._duration}, "
            f"area: {self._area:.3g}, beta: {self._beta:.3g})"
        )

    def __mul__(self, other: float) -> KaiserWaveform:
        return KaiserWaveform(
            self._duration, self._area * float(other), self._beta
        )


# To replicate __init__'s signature in __new__ for every Waveform subclass
def _copy_func(f: FunctionType) -> FunctionType:
    return FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )


for m in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    if m[1].__module__ == __name__:
        _new = _copy_func(m[1].__new__)
        m[1].__new__ = functools.update_wrapper(_new, m[1].__init__)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import pulser
from pulser.channels import Raman, Rydberg


def test_device_channels():
    for dev in pulser.devices._valid_devices:
        for i, (id, ch) in enumerate(dev.channels.items()):
            assert id == dev._channels[i][0]
            assert isinstance(id, str)
            assert ch == dev._channels[i][1]
            assert isinstance(ch, pulser.channels.Channel)
            assert ch.name in ["Rydberg", "Raman"]
            assert ch.basis in ["digital", "ground-rydberg"]
            assert ch.addressing in ["Local", "Global"]
            assert ch.max_abs_detuning >= 0
            assert ch.max_amp > 0
            assert ch.clock_period >= 1
            assert ch.min_duration >= 1
            if ch.addressing == "Local":
                assert ch.retarget_time >= 0
                assert ch.retarget_time == int(ch.retarget_time)
                assert ch.max_targets >= 1
                assert ch.max_targets == int(ch.max_targets)


def test_validate_duration():
    ch = Rydberg.Local(20, 10, min_duration=16, max_duration=1000)
    with pytest.raises(TypeError, match="castable to an int"):
        ch.validate_duration("twenty")
    with pytest.raises(ValueError, match="at least 16 ns"):
        ch.validate_duration(10)
    with pytest.raises(ValueError, match="at most 1000 ns"):
        ch.validate_duration(1e5)
    with pytest.warns(UserWarning, match="not a multiple"):
        ch.validate_duration(31.4)


def test_repr():
    raman = Raman.Local(10, 2, retarget_time=1000, max_targets=4)
    r1 = (
        "Raman.Local(Max Absolute Detuning: 10 rad/µs, Max Amplitude: "
        "2 rad/µs, Target time: 1000 ns, Max targets: 4, Basis: 'digital')"
    )
    assert raman.__str__() == r1

    ryd = Rydberg.Global(50, 2.5)
    r2 = (
        "Rydberg.Global(Max Absolute Detuning: 50 rad/µs, "
        "Max Amplitude: 2.5 rad/µs, Basis: 'ground-rydberg')"
    )
    assert ryd.__str__() == r2



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import FrozenInstanceError

import numpy as np
import pytest

from pulser import Pulse
from pulser.parametrized import Variable
from pulser.waveforms import BlackmanWaveform, CompositeWaveform


a = Variable("a", float)
b = Variable("b", int, size=2)
b._assign([-1.5, 1.5])
c = Variable("c", str)
t = Variable("t", int)
bwf = BlackmanWaveform(t, a)
pulse = Pulse.ConstantDetuning(bwf, b[0], b[1])
pulse2 = Pulse(bwf, bwf, 1)


def test_var():
    with pytest.raises(TypeError, match="'name' has to be of type 'str'"):
        Variable(1, dtype=int)
    with pytest.raises(TypeError, match="Invalid data type"):
        Variable("x", dtype=list, size=4)
    with pytest.raises(TypeError, match="'size' is not of type 'int'"):
        Variable("x", dtype=float, size=(2, 2))
    with pytest.raises(ValueError, match="size 1 or larger"):
        Variable("x", dtype=str, size=0)
    x = Variable("x", dtype=float)
    assert x.value is None
    assert x._count == 0
    with pytest.raises(FrozenInstanceError):
        x.value = 0.5

    assert a.variables == {"a": a}
    assert len(b) == 2
    assert str(c) == "c"

    with pytest.raises(ValueError, match="to variable of size 2"):
        b._assign([1, 4, 5])
    assert np.all(b.build() == np.array([-1, 1]))
    assert b._count == 1

    with pytest.raises(ValueError, match="string to float"):
        a._assign("something")
    with pytest.raises(ValueError, match="No value"):
        a.build()

    with pytest.raises(TypeError, match="must be of type 'str'"):
        c._assign(3.14)

    d = Variable("d", str, size=2)
    d._assign(["o", "k"])
    assert np.all(d.build() == np.array(["o", "k"]))

    with pytest.raises(TypeError, match="Invalid key type"):
        b[[0, 1]]
    with pytest.raises(TypeError, match="not subscriptable"):
        a[0]
    with pytest.raises(IndexError):
        b[2]


def test_varitem():
    b1 = b[1]
    b01 = b[100::-1]
    assert b01.variables == {"b": b}
    assert str(b1) == "b[1]"
    assert str(b01) == "b[100::-1]"
    assert b1.build() == 1
    assert np.all(b01.build() == np.array([1, -1]))
    with pytest.raises(FrozenInstanceError):
        b1.key = 0


def test_paramobj():
    assert set(bwf.variables.keys()) == {"t", "a"}
    assert set(pulse.variables.keys()) == {"t", "a", "b"}
    assert str(bwf) == "BlackmanWaveform(t, a)"
    assert str(pulse) == f"Pulse.ConstantDetuning({str(bwf)}, b[0], b[1])"
    assert str(pulse2) == f"Pulse({str(bwf)}, {str(bwf)}, 1)"
    with pytest.raises(AttributeError):
        bwf._duration
    time = bwf.duration
    samps = bwf.samples
    cwf = CompositeWaveform(bwf, bwf)
    t._assign(1000)
    a._assign(np.pi)
    assert len(cwf.build().samples) == len(samps.build()) * 2
    assert time.build() == 1000


def test_opsupport():
    a._assign(-2.0)
    x = 5 + a
    x = b - x  # x = [-4, -2]
    x = x / 2
    x = 8 * x  # x = [-16, -8]
    x = -x // 3  # x = [5, 2]
    assert np.all(x.build() == [5.0, 2.0])
    assert (a ** a).build() == 0.25
    assert abs(a).build() == 2.0
    assert (3 % a).build() == -1.0



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the configuration of an array of neutral atoms."""

from __future__ import annotations

from collections.abc import Mapping, Iterable
import matplotlib.pyplot as plt
from matplotlib import collections as mc
import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial import KDTree
from typing import Any, cast, Optional, Union

import pulser
from pulser.json.utils import obj_to_dict

QubitId = Union[int, str]


class Register:
    """A quantum register containing a set of qubits.

    Args:
        qubits (dict): Dictionary with the qubit names as keys and their
            position coordinates (in μm) as values
            (e.g. {'q0':(2, -1, 0), 'q1':(-5, 10, 0), ...}).
    """

    def __init__(self, qubits: Mapping[Any, ArrayLike]):
        """Initializes a custom Register."""
        if not isinstance(qubits, dict):
            raise TypeError(
                "The qubits have to be stored in a dictionary "
                "matching qubit ids to position coordinates."
            )
        if not qubits:
            raise ValueError(
                "Cannot create a Register with an empty qubit " "dictionary."
            )
        self._ids = list(qubits.keys())
        coords = [np.array(v, dtype=float) for v in qubits.values()]
        self._dim = coords[0].size
        if any(c.shape != (self._dim,) for c in coords) or (
            self._dim != 2 and self._dim != 3
        ):
            raise ValueError(
                "All coordinates must be specified as vectors of"
                " size 2 or 3."
            )
        self._coords = coords

    @property
    def qubits(self) -> dict[QubitId, np.ndarray]:
        """Dictionary of the qubit names and their position coordinates."""
        return dict(zip(self._ids, self._coords))

    @classmethod
    def from_coordinates(
        cls,
        coords: np.ndarray,
        center: bool = True,
        prefix: Optional[str] = None,
    ) -> Register:
        """Creates the register from an array of coordinates.

        Args:
            coords (ndarray): The coordinates of each qubit to include in the
                register.

        Keyword args:
            center(defaut=True): Whether or not to center the entire array
                around the origin.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed on the given coordinates.
        """
        if center:
            coords = coords - np.mean(coords, axis=0)  # Centers the array
        if prefix is not None:
            pre = str(prefix)
            qubits = {pre + str(i): pos for i, pos in enumerate(coords)}
        else:
            qubits = dict(cast(Iterable, enumerate(coords)))
        return cls(qubits)

    @classmethod
    def rectangle(
        cls,
        rows: int,
        columns: int,
        spacing: float = 4.0,
        prefix: Optional[str] = None,
    ) -> Register:
        """Initializes the register with the qubits in a rectangular array.

        Args:
            rows (int): Number of rows.
            columns (int): Number of columns.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...)

        Returns:
            Register: A register with qubits placed in a rectangular array.
        """
        # Check rows
        if rows < 1:
            raise ValueError(
                f"The number of rows (`rows` = {rows})"
                " must be greater than or equal to 1."
            )

        # Check columns
        if columns < 1:
            raise ValueError(
                f"The number of columns (`columns` = {columns})"
                " must be greater than or equal to 1."
            )

        # Check spacing
        if spacing <= 0.0:
            raise ValueError(
                f"Spacing between atoms (`spacing` = {spacing})"
                " must be greater than 0."
            )

        coords = (
            np.array(
                [(x, y) for y in range(rows) for x in range(columns)],
                dtype=float,
            )
            * spacing
        )

        return cls.from_coordinates(coords, center=True, prefix=prefix)

    @classmethod
    def square(
        cls, side: int, spacing: float = 4.0, prefix: Optional[str] = None
    ) -> Register:
        """Initializes the register with the qubits in a square array.

        Args:
            side (int): Side of the square in number of qubits.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed in a square array.
        """
        # Check side
        if side < 1:
            raise ValueError(
                f"The number of atoms per side (`side` = {side})"
                " must be greater than or equal to 1."
            )

        return cls.rectangle(side, side, spacing=spacing, prefix=prefix)

    @classmethod
    def triangular_lattice(
        cls,
        rows: int,
        atoms_per_row: int,
        spacing: float = 4.0,
        prefix: Optional[str] = None,
    ) -> Register:
        """Initializes the register with the qubits in a triangular lattice.

        Initializes the qubits in a triangular lattice pattern, more
        specifically a triangular lattice with horizontal rows, meaning the
        triangles are pointing up and down.

        Args:
            rows (int): Number of rows.
            atoms_per_row (int): Number of atoms per row.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed in a triangular lattice.
        """
        # Check rows
        if rows < 1:
            raise ValueError(
                f"The number of rows (`rows` = {rows})"
                " must be greater than or equal to 1."
            )

        # Check atoms per row
        if atoms_per_row < 1:
            raise ValueError(
                "The number of atoms per row"
                f" (`atoms_per_row` = {atoms_per_row})"
                " must be greater than or equal to 1."
            )

        # Check spacing
        if spacing <= 0.0:
            raise ValueError(
                f"Spacing between atoms (`spacing` = {spacing})"
                " must be greater than 0."
            )

        coords = np.array(
            [(x, y) for y in range(rows) for x in range(atoms_per_row)],
            dtype=float,
        )
        coords[:, 0] += 0.5 * np.mod(coords[:, 1], 2)
        coords[:, 1] *= np.sqrt(3) / 2
        coords *= spacing

        return cls.from_coordinates(coords, center=True, prefix=prefix)

    @classmethod
    def _hexagon_helper(
        cls,
        layers: int,
        atoms_left: int,
        spacing: float,
        prefix: Optional[str] = None,
    ) -> Register:
        """Helper function for building hexagonal arrays.

        Args:
            layers (int): Number of full layers around a central atom.
            atoms_left (int): Number of atoms on the external layer.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed in a hexagonal layout
                with extra atoms on the outermost layer if needed.
        """
        # y coordinates of the top vertex of a triangle
        crest_y = np.sqrt(3) / 2.0

        # Coordinates of vertices
        start_x = [-1.0, -0.5, 0.5, 1.0, 0.5, -0.5]
        start_y = [0.0, crest_y, crest_y, 0, -crest_y, -crest_y]

        # Steps to place atoms, starting from a vertex
        delta_x = [0.5, 1.0, 0.5, -0.5, -1.0, -0.5]
        delta_y = [crest_y, 0.0, -crest_y, -crest_y, 0.0, crest_y]

        coords = np.array(
            [
                (
                    start_x[side] * layer + atom * delta_x[side],
                    start_y[side] * layer + atom * delta_y[side],
                )
                for layer in range(1, layers + 1)
                for side in range(6)
                for atom in range(1, layer + 1)
            ],
            dtype=float,
        )

        if atoms_left > 0:
            layer = layers + 1
            min_atoms_per_side = atoms_left // 6
            # Extra atoms after balancing all sides
            atoms_left %= 6

            # Order for placing left atoms
            # Top-Left, Top-Right, Bottom (C3 symmetry)...
            # ...Top, Bottom-Right, Bottom-Left (C6 symmetry)
            sides_order = [0, 3, 1, 4, 2, 5]

            coords2 = np.array(
                [
                    (
                        start_x[side] * layer + atom * delta_x[side],
                        start_y[side] * layer + atom * delta_y[side],
                    )
                    for side in range(6)
                    for atom in range(
                        1,
                        min_atoms_per_side + 2
                        if atoms_left > sides_order[side]
                        else min_atoms_per_side + 1,
                    )
                ],
                dtype=float,
            )

            coords = np.concatenate((coords, coords2))

        coords *= spacing
        coords = np.concatenate(([(0.0, 0.0)], coords))

        return cls.from_coordinates(coords, center=False, prefix=prefix)

    @classmethod
    def hexagon(
        cls, layers: int, spacing: float = 4.0, prefix: Optional[str] = None
    ) -> Register:
        """Initializes the register with the qubits in a hexagonal layout.

        Args:
            layers (int): Number of layers around a central atom.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed in a hexagonal layout.
        """
        # Check layers
        if layers < 1:
            raise ValueError(
                f"The number of layers (`layers` = {layers})"
                " must be greater than or equal to 1."
            )

        # Check spacing
        if spacing <= 0.0:
            raise ValueError(
                f"Spacing between atoms (`spacing` = {spacing})"
                " must be greater than 0."
            )

        return cls._hexagon_helper(layers, 0, spacing, prefix)

    @classmethod
    def max_connectivity(
        cls,
        n_qubits: int,
        device: pulser.devices._device_datacls.Device,
        spacing: float = None,
        prefix: str = None,
    ) -> Register:
        """Initializes the register with maximum connectivity for a given device.

        In order to maximize connectivity, the basic pattern is the triangle.
        Atoms are first arranged as layers of hexagons around a central atom.
        Extra atoms are placed in such a manner that C3 and C6 rotational
        symmetries are enforced as often as possible.

        Args:
            n_qubits (int): Number of qubits.
            device (Device): The device whose constraints must be obeyed.

        Keyword args:
            spacing(float): The distance between neighbouring qubits in μm.
                If omitted, the minimal distance for the device is used.
            prefix (str): The prefix for the qubit ids. If defined, each qubit
                id starts with the prefix, followed by an int from 0 to N-1
                (e.g. prefix='q' -> IDs: 'q0', 'q1', 'q2', ...).

        Returns:
            Register: A register with qubits placed for maximum connectivity.
        """
        # Check device
        if not isinstance(device, pulser.devices._device_datacls.Device):
            raise TypeError(
                "'device' must be of type 'Device'. Import a valid"
                " device from 'pulser.devices'."
            )

        # Check number of qubits (1 or above)
        if n_qubits < 1:
            raise ValueError(
                f"The number of qubits (`n_qubits` = {n_qubits})"
                " must be greater than or equal to 1."
            )

        # Check number of qubits (less than the max number of atoms)
        if n_qubits > device.max_atom_num:
            raise ValueError(
                f"The number of qubits (`n_qubits` = {n_qubits})"
                " must be less than or equal to the maximum"
                " number of atoms supported by this device"
                f" ({device.max_atom_num})."
            )

        # Default spacing or check minimal distance
        if spacing is None:
            spacing = device.min_atom_distance
        elif spacing < device.min_atom_distance:
            raise ValueError(
                f"Spacing between atoms (`spacing = `{spacing})"
                " must be greater than or equal to the minimal"
                " distance supported by this device"
                f" ({device.min_atom_distance})."
            )

        if n_qubits < 7:
            crest_y = np.sqrt(3) / 2.0
            hex_coords = np.array(
                [
                    (0.0, 0.0),
                    (-0.5, crest_y),
                    (0.5, crest_y),
                    (1.0, 0.0),
                    (0.5, -crest_y),
                    (-0.5, -crest_y),
                ]
            )
            return cls.from_coordinates(
                spacing * hex_coords[:n_qubits], prefix=prefix, center=False
            )

        full_layers = int((-3.0 + np.sqrt(9 + 12 * (n_qubits - 1))) / 6.0)
        atoms_left = n_qubits - 1 - (full_layers ** 2 + full_layers) * 3

        return cls._hexagon_helper(full_layers, atoms_left, spacing, prefix)

    def rotate(self, degrees: float) -> None:
        """Rotates the array around the origin by the given angle.

        Args:
            degrees (float): The angle of rotation in degrees.
        """
        if self._dim != 2:
            raise NotImplementedError("Can only rotate arrays in 2D.")
        theta = np.deg2rad(degrees)
        rot = np.array(
            [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
        )
        self._coords = [rot @ v for v in self._coords]

    def draw(
        self,
        with_labels: bool = True,
        blockade_radius: Optional[float] = None,
        draw_graph: bool = True,
        draw_half_radius: bool = False,
    ) -> None:
        """Draws the entire register.

        Keyword Args:
            with_labels(bool, default=True): If True, writes the qubit ID's
                next to each qubit.
            blockade_radius(float, default=None): The distance (in μm) between
                atoms below the Rydberg blockade effect occurs.
            draw_half_radius(bool, default=False): Whether or not to draw the
                half the blockade radius surrounding each atoms. If `True`,
                requires `blockade_radius` to be defined.
            draw_graph(bool, default=True): Whether or not to draw the
                interaction between atoms as edges in a graph. Will only draw
                if the `blockade_radius` is defined.

        Note:
            When drawing half the blockade radius, we say there is a blockade
            effect between atoms whenever their respective circles overlap.
            This representation is preferred over drawing the full Rydberg
            radius because it helps in seeing the interactions between atoms.
        """
        # Check dimensions
        if self._dim != 2:
            raise NotImplementedError("Can only draw register layouts in 2D.")

        # Check spacing
        if blockade_radius is not None and blockade_radius <= 0.0:
            raise ValueError(
                "Blockade radius (`blockade_radius` ="
                f" {blockade_radius})"
                " must be greater than 0."
            )

        pos = np.array(self._coords)
        diffs = np.max(pos, axis=0) - np.min(pos, axis=0)
        diffs[diffs < 9] *= 1.5
        diffs[diffs < 9] += 2
        if blockade_radius and draw_half_radius:
            diffs[diffs < blockade_radius] = blockade_radius
        big_side = max(diffs)
        proportions = diffs / big_side
        Ls = proportions * min(
            big_side / 4, 10
        )  # Figsize is, at most, (10,10)

        fig, ax = plt.subplots(figsize=Ls)
        ax.scatter(pos[:, 0], pos[:, 1], s=30, alpha=0.7, c="darkgreen")

        ax.set_xlabel("µm")
        ax.set_ylabel("µm")
        ax.axis("equal")
        ax.spines["right"].set_color("none")
        ax.spines["top"].set_color("none")

        if with_labels:
            for q, coords in zip(self._ids, self._coords):
                ax.annotate(q, coords, fontsize=12, ha="left", va="bottom")

        if draw_half_radius:
            if blockade_radius is None:
                raise ValueError("Define 'blockade_radius' to draw.")
            if len(pos) == 1:
                raise NotImplementedError(
                    "Needs more than one atom to draw " "the blockade radius."
                )

            for p in pos:
                circle = plt.Circle(
                    tuple(p), blockade_radius / 2, alpha=0.1, color="darkgreen"
                )
                ax.add_patch(circle)
        if draw_graph and blockade_radius is not None:
            epsilon = 1e-9  # Accounts for rounding errors
            edges = KDTree(pos).query_pairs(blockade_radius * (1 + epsilon))
            lines = pos[(tuple(edges),)]
            lc = mc.LineCollection(lines, linewidths=0.6, colors="grey")
            ax.add_collection(lc)

        else:
            # Only draw central axis lines when not drawing the graph
            ax.axvline(0, c="grey", alpha=0.5, linestyle=":")
            ax.axhline(0, c="grey", alpha=0.5, linestyle=":")

        plt.show()

    def _to_dict(self) -> dict[str, Any]:
        qs = dict(zip(self._ids, map(np.ndarray.tolist, self._coords)))
        return obj_to_dict(self, qs)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the Variable and auxiliary classes."""

from __future__ import annotations

import collections.abc  # To use collections.abc.Sequence
from collections.abc import Iterable
import dataclasses
from typing import Union, Any, cast

import numpy as np
from numpy.typing import ArrayLike

from pulser.parametrized import Parametrized
from pulser.parametrized.paramobj import OpSupport
from pulser.json.utils import obj_to_dict


@dataclasses.dataclass(frozen=True, eq=False)
class Variable(Parametrized, OpSupport):
    """A variable for parametrized sequence building.

    Args:
        name (str): Unique name for the variable.
        dtype (type): Type of the variable's content. Supports `float`, `int`
            and `str`.
        size (int=1): The number of values stored. Defaults to a single value.
    """

    name: str
    dtype: Union[type[float], type[int], type[str]]
    size: int = 1

    def __post_init__(self) -> None:
        if not isinstance(self.name, str):
            raise TypeError("Variable's 'name' has to be of type 'str'.")
        if self.dtype not in [int, float, str]:
            raise TypeError(f"Invalid data type '{self.dtype}' for Variable.")
        if not isinstance(self.size, int):
            raise TypeError("Given variable 'size' is not of type 'int'.")
        elif self.size < 1:
            raise ValueError("Variables must be of size 1 or larger.")

        self._count: int
        self.__dict__["_count"] = -1  # Counts the updates
        self._clear()

    @property
    def variables(self) -> dict[str, Variable]:
        """Returns a dictionary with the only variable involved (itself)."""
        return {self.name: self}

    def _clear(self) -> None:
        self.__dict__["value"] = None
        self.__dict__["_count"] += 1

    def _assign(self, value: Union[ArrayLike, str, float, int]) -> None:
        if self.dtype == str:
            if not (
                isinstance(value, str)
                if self.size == 1
                else all(isinstance(s, str) for s in cast(Iterable, value))
            ):
                raise TypeError(
                    f"Provided values for variable '{self.name}' "
                    "must be of type 'str'."
                )

        val = np.array(value, dtype=self.dtype)
        if val.size != self.size:
            raise ValueError(
                f"Can't assign array of size {val.size} to "
                + f"variable of size {self.size}."
            )

        self.__dict__["value"] = self.dtype(val) if self.size == 1 else val
        self.__dict__["_count"] += 1

    def build(self) -> Union[ArrayLike, str, float, int]:
        """Returns the variable's current value."""
        self.value: Union[ArrayLike, str, float, int]
        if self.value is None:
            raise ValueError(f"No value assigned to variable '{self.name}'.")
        return self.value

    def _to_dict(self) -> dict[str, Any]:
        d = obj_to_dict(self, _build=False)
        d.update(dataclasses.asdict(self))
        return d

    def __str__(self) -> str:
        return self.name

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, key: Union[int, slice]) -> _VariableItem:
        if not isinstance(key, (int, slice)):
            raise TypeError(f"Invalid key type {type(key)} for '{self.name}'.")
        if self.size == 1:
            raise TypeError(f"Variable '{self.name}' is not subscriptable.")
        if isinstance(key, int):
            if not -self.size <= key < self.size:
                raise IndexError(f"{key} outside of range for '{self.name}'.")

        return _VariableItem(self, key)


@dataclasses.dataclass(frozen=True)
class _VariableItem(Parametrized, OpSupport):
    """Stores access to items of a variable with multiple values."""

    var: Variable
    key: Union[int, slice]

    @property
    def variables(self) -> dict[str, Variable]:
        return self.var.variables

    def build(self) -> Union[ArrayLike, str, float, int]:
        """Return the variable's item(s) values."""
        return cast(collections.abc.Sequence, self.var.build())[self.key]

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(
            self, self.var, self.key, _module="operator", _name="getitem"
        )

    def __str__(self) -> str:
        if isinstance(self.key, slice):
            items = [
                "" if x is None else str(x)
                for x in [self.key.start, self.key.stop, self.key.step]
            ]
            key_str = ":".join(items)
        else:
            key_str = str(self.key)
        return f"{str(self.var)}[{key_str}]"



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import numpy as np
import pytest

from pulser import Pulse
from pulser.waveforms import ConstantWaveform, BlackmanWaveform, RampWaveform

cwf = ConstantWaveform(100, -10)
bwf = BlackmanWaveform(200, 3)
rwf = RampWaveform(200, 0, 1)

pls = Pulse(bwf, bwf, 2 * np.pi)
pls2 = Pulse.ConstantPulse(100, 1, -10, -np.pi)
pls3 = Pulse.ConstantAmplitude(1, cwf, 1)
pls4 = Pulse.ConstantDetuning(bwf, -10, 0)


def test_creation():
    with pytest.raises(TypeError):
        Pulse(10, 0, 0, post_phase_shift=2)
        Pulse(cwf, 1, 0)
        Pulse(0, bwf, 1)
        Pulse(bwf, cwf, bwf)
        Pulse(bwf, cwf, 0, post_phase_shift=cwf)

    with pytest.raises(ValueError, match="The duration of"):
        Pulse(bwf, cwf, 0)

    with pytest.raises(ValueError, match="All samples of an amplitude"):
        Pulse(cwf, cwf, 0)
        Pulse.ConstantAmplitude(-1, cwf, 0)
        Pulse.ConstantPulse(100, -1, 0, 0)

    assert pls.phase == 0
    assert pls2.amplitude == pls3.amplitude
    assert pls2.detuning == pls3.detuning
    assert pls2.phase == np.pi
    assert pls3.phase == 1
    assert pls4.detuning != cwf
    assert pls4.amplitude == pls.amplitude


def test_str():
    assert pls2.__str__() == (
        "Pulse(Amp=1 rad/µs, Detuning=-10 rad/µs, " + "Phase=3.14)"
    )
    pls_ = Pulse(bwf, rwf, 1)
    msg = "Pulse(Amp=Blackman(Area: 3), Detuning=Ramp(0->1 rad/µs), Phase=1)"
    assert pls_.__str__() == msg


def test_repr():
    pls_ = Pulse(bwf, rwf, 1, post_phase_shift=-np.pi)
    msg = (
        "Pulse(amp=BlackmanWaveform(200 ns, Area: 3), "
        + "detuning=RampWaveform(200 ns, 0->1 rad/µs), "
        + "phase=1, post_phase_shift=3.14)"
    )
    assert pls_.__repr__() == msg


def test_draw():
    pls_ = Pulse.ConstantDetuning(bwf, -10, 1, post_phase_shift=-np.pi)
    with patch("matplotlib.pyplot.show"):
        pls_.draw()



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from unittest.mock import patch

import numpy as np
import pytest

from pulser import Sequence, Register
from pulser.devices import Chadoq2
from pulser.json.coders import PulserEncoder, PulserDecoder
from pulser.parametrized.decorators import parametrize
from pulser.waveforms import BlackmanWaveform


def encode(obj):
    return json.dumps(obj, cls=PulserEncoder)


def decode(s):
    return json.loads(s, cls=PulserDecoder)


def encode_decode(obj):
    return decode(encode(obj))


def test_encoder():
    assert np.all(np.arange(10) == encode_decode(np.arange(10)))
    assert set(range(5)) == encode_decode(set(range(5)))
    with pytest.raises(TypeError, match="not JSON serializable"):
        encode(1j)


def test_rare_cases():
    reg = Register.square(4)
    seq = Sequence(reg, Chadoq2)
    var = seq.declare_variable("var")

    wf = BlackmanWaveform(100, var)
    with pytest.warns(
        UserWarning, match="Calls to methods of parametrized " "objects"
    ):
        s = encode(wf.draw())

    with pytest.warns(UserWarning, match="not encode a Sequence"):
        wf_ = Sequence.deserialize(s)

    var._assign(-10)
    with pytest.raises(ValueError, match="No value assigned"):
        wf_.build()

    var_ = wf_._variables["var"]
    var_._assign(-10)
    with patch("matplotlib.pyplot.show"):
        wf_.build()

    rotated_reg = parametrize(Register.rotate)(reg, var)
    with pytest.raises(NotImplementedError):
        encode(rotated_reg)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import FrozenInstanceError
from unittest.mock import patch

import numpy as np
import pytest

import pulser
from pulser.devices import Chadoq2
from pulser.register import Register


def test_init():
    for dev in pulser.devices._valid_devices:
        assert dev.dimensions in (2, 3)
        assert dev.max_atom_num > 10
        assert dev.max_radial_distance > 10
        assert dev.min_atom_distance > 0
        assert dev.interaction_coeff > 0
        assert isinstance(dev.channels, dict)
        with pytest.raises(FrozenInstanceError):
            dev.name = "something else"
    assert Chadoq2 in pulser.devices._valid_devices
    assert Chadoq2.supported_bases == {"digital", "ground-rydberg"}
    with patch("sys.stdout"):
        Chadoq2.print_specs()
    assert Chadoq2.__repr__() == "Chadoq2"


def test_mock():
    dev = pulser.devices.MockDevice
    assert dev.dimensions == 2
    assert dev.max_atom_num > 1000
    assert dev.min_atom_distance <= 1
    assert dev.interaction_coeff == 5008713
    names = ["Rydberg", "Raman", "Microwave"]
    basis = ["ground-rydberg", "digital", "XY"]
    for ch in dev.channels.values():
        assert ch.name in names
        assert ch.basis == basis[names.index(ch.name)]
        assert ch.addressing in ["Local", "Global"]
        assert ch.max_abs_detuning >= 1000
        assert ch.max_amp >= 200
        if ch.addressing == "Local":
            assert ch.retarget_time == 0
            assert ch.max_targets > 1
            assert ch.max_targets == int(ch.max_targets)


def test_rydberg_blockade():
    dev = pulser.devices.MockDevice
    assert np.isclose(dev.rydberg_blockade_radius(3 * np.pi), 9)
    assert np.isclose(dev.rabi_from_blockade(9), 3 * np.pi)
    rand_omega = np.random.rand() * 2 * np.pi
    assert np.isclose(
        rand_omega,
        dev.rabi_from_blockade(dev.rydberg_blockade_radius(rand_omega)),
    )


def test_validate_register():
    with pytest.raises(ValueError, match="The number of atoms"):
        Chadoq2.validate_register(Register.square(50))

    coords = [(100, 0), (-100, 0)]
    with pytest.raises(TypeError):
        Chadoq2.validate_register(coords)
    with pytest.raises(ValueError, match="at most 50 μm away from the center"):
        Chadoq2.validate_register(Register.from_coordinates(coords))

    with pytest.raises(ValueError, match="must be 2D vectors"):
        coords = [(-10, 4, 0), (0, 0, 0)]
        Chadoq2.validate_register(Register(dict(enumerate(coords))))

    with pytest.raises(ValueError, match="don't respect the minimal distance"):
        Chadoq2.validate_register(
            Register.triangular_lattice(3, 4, spacing=3.9)
        )

    Chadoq2.validate_register(Register.rectangle(5, 10, spacing=5))



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Sequence class, where a pulse sequence is defined."""

from __future__ import annotations

from collections import namedtuple
from collections.abc import Callable, Generator, Iterable, Set
import copy
from functools import wraps
from itertools import chain
import json
from sys import version_info
from typing import Any, cast, NamedTuple, Optional, Tuple, Union
import warnings

import numpy as np
from numpy.typing import ArrayLike

import pulser
from pulser.channels import Channel
from pulser.devices import MockDevice
from pulser.devices._device_datacls import Device
from pulser.json.coders import PulserEncoder, PulserDecoder
from pulser.json.utils import obj_to_dict
from pulser.parametrized import Parametrized, Variable
from pulser.pulse import Pulse
from pulser.register import Register
from pulser._seq_drawer import draw_sequence

if version_info[:2] >= (3, 8):  # pragma: no cover
    from typing import Literal, get_args
else:  # pragma: no cover
    try:
        from typing_extensions import Literal, get_args  # type: ignore
    except ImportError:
        raise ImportError(
            "Using pulser with Python version 3.7 requires the"
            " `typing_extensions` module. Install it by running"
            " `pip install typing-extensions`."
        )


QubitId = Union[int, str]
PROTOCOLS = Literal["min-delay", "no-delay", "wait-for-all"]


class _TimeSlot(NamedTuple):
    """Auxiliary class to store the information in the schedule."""

    type: Union[Pulse, str]
    ti: int
    tf: int
    targets: Set[QubitId]


# Encodes a sequence building calls
_Call = namedtuple("_Call", ["name", "args", "kwargs"])


def _screen(func: Callable) -> Callable:
    """Blocks the call to a function if the Sequence is parametrized."""

    @wraps(func)
    def wrapper(self: Sequence, *args: Any, **kwargs: Any) -> Any:
        if self.is_parametrized():
            raise RuntimeError(
                f"Sequence.{func.__name__} can't be called in"
                " parametrized sequences."
            )
        return func(self, *args, **kwargs)

    return wrapper


def _store(func: Callable) -> Callable:
    """Stores any Sequence building call for deferred execution."""

    @wraps(func)
    def wrapper(self: Sequence, *args: Any, **kwargs: Any) -> Any:
        def verify_variable(x: Any) -> None:
            if isinstance(x, Parametrized):
                # If not already, the sequence becomes parametrized
                self._building = False
                for name, var in x.variables.items():
                    if name not in self._variables:
                        raise ValueError(f"Unknown variable '{name}'.")
                    elif self._variables[name] is not var:
                        raise ValueError(
                            f"{x} has variables that don't come from this "
                            "Sequence. Use only what's returned by this"
                            "Sequence's 'declare_variable' method as your"
                            "variables."
                        )
            elif isinstance(x, Iterable) and not isinstance(x, str):
                # Recursively look for parametrized objs inside the arguments
                for y in x:
                    verify_variable(y)

        if self._is_measured and self.is_parametrized():
            raise RuntimeError(
                "The sequence has been measured, no further "
                "changes are allowed."
            )
        # Check if all Parametrized inputs stem from declared variables
        for x in chain(args, kwargs.values()):
            verify_variable(x)
        storage = self._calls if self._building else self._to_build_calls
        func(self, *args, **kwargs)
        storage.append(_Call(func.__name__, args, kwargs))

    return wrapper


class Sequence:
    """A sequence of operations on a device.

    A sequence is composed by

        - The device in which we want to implement it
        - The register of qubits on which to act
        - The device's channels that are used
        - The schedule of operations on each channel


    A Sequence also supports variable parameters, which have to be obtained
    through ``Sequence.declare_variable()``. From the moment a variable is
    declared, a ``Sequence`` becomes **parametrized** and stops being built on
    the fly, instead storing the sequence building calls for later execution.
    This forgoes some specific functionalities of a "regular" ``Sequence``,
    like the ability to validate a ``Pulse`` or to draw the sequence as it is
    being built. Instead, all validation happens upon building (through
    ``Sequence.build()``), where values for all declared variables have to be
    specified and a "regular" ``Sequence`` is created and returned. By
    changing the values given to the variables, multiple sequences can be
    generated from a single "parametrized" ``Sequence``.

    Args:
        register(Register): The atom register on which to apply the pulses.
        device(Device): A valid device in which to execute the Sequence (import
            it from ``pulser.devices``).

    Note:
        The register and device do not support variable parameters. As such,
        they are the same for all Sequences built from a parametrized Sequence.
    """

    def __init__(self, register: Register, device: Device):
        """Initializes a new pulse sequence."""
        if not isinstance(device, Device):
            raise TypeError(
                "'device' must be of type 'Device'. Import a valid"
                " device from 'pulser.devices'."
            )
        cond1 = device not in pulser.devices._valid_devices
        cond2 = device != MockDevice
        if cond1 and cond2:
            names = [d.name for d in pulser.devices._valid_devices]
            warns_msg = (
                "The Sequence's device should be imported from "
                + "'pulser.devices'. Correct operation is not ensured"
                + " for custom devices. Choose 'MockDevice' or one of"
                + " the following real devices:\n"
                + "\n".join(names)
            )
            warnings.warn(warns_msg, stacklevel=2)

        # Checks if register is compatible with the device
        device.validate_register(register)

        self._register: Register = register
        self._device: Device = device
        self._in_xy: bool = False
        self._calls: list[_Call] = [_Call("__init__", (register, device), {})]
        self._channels: dict[str, Channel] = {}
        self._schedule: dict[str, list[_TimeSlot]] = {}
        # The phase reference of each channel
        self._phase_ref: dict[str, dict[QubitId, _PhaseTracker]] = {}
        # Stores the names and dict ids of declared channels
        self._taken_channels: dict[str, str] = {}
        # IDs of all qubits in device
        self._qids: set[QubitId] = set(self.qubit_info.keys())
        # Last time each qubit was used, by basis
        self._last_used: dict[str, dict[QubitId, int]] = {}
        # Last time a target happened, by channel
        self._last_target: dict[str, int] = {}
        self._variables: dict[str, Variable] = {}
        self._to_build_calls: list[_Call] = []
        self._building: bool = True

        # Initializes all parametrized Sequence related attributes
        self._reset_parametrized()

    @property
    def qubit_info(self) -> dict[QubitId, np.ndarray]:
        """Dictionary with the qubit's IDs and positions."""
        return self._register.qubits

    @property
    def declared_channels(self) -> dict[str, Channel]:
        """Channels declared in this Sequence."""
        return dict(self._channels)

    @property
    def declared_variables(self) -> dict[str, Variable]:
        """Variables declared in this Sequence."""
        return dict(self._variables)

    @property
    def available_channels(self) -> dict[str, Channel]:
        """Channels still available for declaration."""
        # Show all channels if none are declared, otherwise filter depending
        # on whether the sequence is working on XY mode
        if not self._channels:
            return dict(self._device.channels)
        else:
            # MockDevice channels can be declared multiple times
            return {
                id: ch
                for id, ch in self._device.channels.items()
                if (
                    id not in self._taken_channels.values()
                    or self._device == MockDevice
                )
                and (ch.basis == "XY" if self._in_xy else ch.basis != "XY")
            }

    def is_parametrized(self) -> bool:
        """States whether the sequence is parametrized.

        A parametrized sequence is one that depends on the values assigned to
        variables declared within it. Sequence-building calls are not executed
        right away, but rather stored for deferred execution when all variables
        are given a value (when ``Sequence.build()`` is called).

        Returns:
            bool: Whether the sequence is parametrized.
        """
        return not self._building

    @_screen
    def get_duration(self, channel: Optional[str] = None) -> int:
        """Returns the current duration of a channel or the whole sequence.

        Keyword Args:
            channel (Optional[str]): A specific channel to return the duration
                of. If left as None, it will return the duration of the whole
                sequence.

        Returns:
            int: The duration of the channel or sequence, in ns.
        """
        if channel is None:
            durations = [
                self._last(ch).tf
                for ch in self._schedule
                if self._schedule[ch]
            ]
            return 0 if not durations else max(durations)

        self._validate_channel(channel)
        return self._last(channel).tf if self._schedule[channel] else 0

    @_screen
    def current_phase_ref(
        self, qubit: QubitId, basis: str = "digital"
    ) -> float:
        """Current phase reference of a specific qubit for a given basis.

        Args:
            qubit (Union[int, str]): The id of the qubit whose phase shift is
                desired.
            basis (str): The basis (i.e. electronic transition) the phase
                reference is associated with. Must correspond to the basis of a
                declared channel.

        Returns:
            float: Current phase reference of 'qubit' in 'basis'.
        """
        if qubit not in self._qids:
            raise ValueError(
                "'qubit' must be the id of a qubit declared in "
                "this sequence's device."
            )

        if basis not in self._phase_ref:
            raise ValueError("No declared channel targets the given 'basis'.")

        return self._phase_ref[basis][qubit].last_phase

    def declare_channel(
        self,
        name: str,
        channel_id: str,
        initial_target: Optional[
            Union[
                Iterable[Union[QubitId, Parametrized]],
                Union[QubitId, Parametrized],
            ]
        ] = None,
    ) -> None:
        """Declares a new channel to the Sequence.

        The first declared channel implicitly defines the sequence's mode of
        operation (i.e. the underlying Hamiltonian). In particular, if the
        first declared channel is of type ``Microwave``, the sequence will work
        in "XY Mode" and will not allow declaration of channels that do not
        address the 'XY' basis. Inversely, declaration of a channel of another
        type will block the declaration of ``Microwave`` channels.

        Note:
            Regular devices only allow a channel to be declared once, but
            ``MockDevice`` channels can be repeatedly declared if needed.

        Args:
            name (str): Unique name for the channel in the sequence.
            channel_id (str): How the channel is identified in the device.
                Consult ``Sequence.available_channels`` to see which channel
                ID's are still available and the associated channel's
                description.
            initial_target (Optional[Union[int, str, Iterable]]): For 'Local'
                addressing channels only. Declares the initial target of the
                channel. If left as None, the initial target will have to be
                set manually as the first addition to this channel.
        """
        if name in self._channels:
            raise ValueError("The given name is already in use.")

        if channel_id not in self._device.channels:
            raise ValueError("No channel %s in the device." % channel_id)

        ch = self._device.channels[channel_id]
        if channel_id not in self.available_channels:
            if self._in_xy and ch.basis != "XY":
                raise ValueError(
                    f"Channel '{ch}' cannot work simultaneously "
                    "with the declared 'Microwave' channel."
                )
            elif not self._in_xy and ch.basis == "XY":
                raise ValueError(
                    "Channel of type 'Microwave' cannot work "
                    "simultaneously with the declared channels."
                )
            else:
                raise ValueError(f"Channel {channel_id} is not available.")

        if ch.basis == "XY" and not self._in_xy:
            self._in_xy = True
        self._channels[name] = ch
        self._taken_channels[name] = channel_id
        self._schedule[name] = []
        self._last_target[name] = 0

        if ch.basis not in self._phase_ref:
            self._phase_ref[ch.basis] = {
                q: _PhaseTracker(0) for q in self._qids
            }
            self._last_used[ch.basis] = {q: 0 for q in self._qids}

        if ch.addressing == "Global":
            self._add_to_schedule(name, _TimeSlot("target", -1, 0, self._qids))
        elif initial_target is not None:
            try:
                cond = any(
                    isinstance(t, Parametrized)
                    for t in cast(Iterable, initial_target)
                )
            except TypeError:
                cond = isinstance(initial_target, Parametrized)
            if cond:
                self._building = False

            if self.is_parametrized():
                # Do not store "initial_target" in a _call when parametrized
                # It is stored as a _to_build_call when target is called
                self.target(initial_target, name)
                initial_target = None
            else:
                # "_target" call is not saved
                self._target(
                    cast(Union[Iterable, QubitId], initial_target), name
                )

        # Manually store the channel declaration as a regular call
        self._calls.append(
            _Call(
                "declare_channel",
                (name, channel_id),
                {"initial_target": initial_target},
            )
        )

    def declare_variable(
        self,
        name: str,
        size: int = 1,
        dtype: Union[type[int], type[float], type[str]] = float,
    ) -> Variable:
        """Declare a new variable within this Sequence.

        The declared variables can be used to create parametrized versions of
        ``Waveform`` and ``Pulse`` objects, which in turn can be added to the
        ``Sequence``. Additionally, simple arithmetic operations involving
        variables are also supported and will return parametrized objects that
        are dependent on the involved variables.

        Args:
            name (str): The name for the variable. Must be unique within a
                Sequence.

        Keyword Args:
            size (int=1): The number of entries stored in the variable.
            dtype (default=float): The type of the data that will be assigned
                to the variable. Must be ``float``, ``int`` or ``str``.

        Returns:
            Variable: The declared Variable instance.

        Note:
            To avoid confusion, it is recommended to store the returned
            Variable instance in a Python variable with the same name.
        """
        if name in self._variables:
            raise ValueError("Name for variable is already being used.")
        var = Variable(name, dtype, size=size)
        self._variables[name] = var
        return var

    @_store
    def add(
        self,
        pulse: Union[Pulse, Parametrized],
        channel: Union[str, Parametrized],
        protocol: PROTOCOLS = "min-delay",
    ) -> None:
        """Adds a pulse to a channel.

        Args:
            pulse (pulser.Pulse): The pulse object to add to the channel.
            channel (str): The channel's name provided when declared.
            protocol (str, default='min-delay'): Stipulates how to deal with
                eventual conflicts with other channels, specifically in terms
                of having multiple channels act on the same target
                simultaneously.

                - ``'min-delay'``: Before adding the pulse, introduces the
                  smallest possible delay that avoids all exisiting conflicts.

                - ``'no-delay'``: Adds the pulse to the channel, regardless of
                  existing conflicts.

                - ``'wait-for-all'``: Before adding the pulse, adds a delay
                  that idles the channel until the end of the other channels'
                  latest pulse.
        """
        pulse = cast(Pulse, pulse)
        channel = cast(str, channel)

        self._validate_channel(channel)

        valid_protocols = get_args(PROTOCOLS)
        if protocol not in valid_protocols:
            raise ValueError(
                f"Invalid protocol '{protocol}', only accepts protocols: "
                + ", ".join(valid_protocols)
            )

        if self.is_parametrized():
            if not isinstance(pulse, Parametrized):
                self._validate_pulse(pulse, channel)
            return

        if not isinstance(pulse, Pulse):
            raise TypeError(
                f"'pulse' must be of type Pulse, not of type {type(pulse)}."
            )

        channel_obj = self._channels[channel]
        _duration = channel_obj.validate_duration(pulse.duration)
        if _duration != pulse.duration:
            try:
                pulse = Pulse(
                    pulse.amplitude.change_duration(_duration),
                    pulse.detuning.change_duration(_duration),
                    pulse.phase,
                    pulse.post_phase_shift,
                )
            except NotImplementedError:
                raise TypeError(
                    "Failed to automatically adjust one of the pulse's "
                    "waveforms to the channel duration constraints. Choose a "
                    "duration that is a multiple of "
                    f"{channel_obj.clock_period} ns."
                )

        self._validate_pulse(pulse, channel)
        last = self._last(channel)
        t0 = last.tf  # Preliminary ti
        basis = channel_obj.basis
        phase_barriers = [
            self._phase_ref[basis][q].last_time for q in last.targets
        ]
        current_max_t = max(t0, *phase_barriers)
        if protocol != "no-delay":
            for ch, seq in self._schedule.items():
                if ch == channel:
                    continue
                for op in self._schedule[ch][::-1]:
                    if op.tf <= current_max_t:
                        break
                    if not isinstance(op.type, Pulse):
                        continue
                    if op.targets & last.targets or protocol == "wait-for-all":
                        current_max_t = op.tf
                        break
        ti = current_max_t
        if ti > t0:
            # Insert a delay
            delay_duration = ti - t0

            # Delay must not be shorter than the min duration for this channel
            min_duration = self._channels[channel].min_duration
            if delay_duration < min_duration:
                ti += min_duration - delay_duration
                delay_duration = min_duration

            self._delay(delay_duration, channel)

        tf = ti + pulse.duration

        prs = {self._phase_ref[basis][q].last_phase for q in last.targets}
        if len(prs) != 1:
            raise ValueError(
                "Cannot do a multiple-target pulse on qubits with different "
                "phase references for the same basis."
            )
        else:
            phase_ref = prs.pop()

        if phase_ref != 0:
            # Has to copy to keep the original pulse intact
            pulse = copy.deepcopy(pulse)
            pulse.phase = (pulse.phase + phase_ref) % (2 * np.pi)

        self._add_to_schedule(channel, _TimeSlot(pulse, ti, tf, last.targets))

        for qubit in last.targets:
            if self._last_used[basis][qubit] < tf:
                self._last_used[basis][qubit] = tf

        if pulse.post_phase_shift:
            self._phase_shift(
                pulse.post_phase_shift, *last.targets, basis=basis
            )

    @_store
    def target(
        self,
        qubits: Union[QubitId, Iterable[QubitId], Parametrized],
        channel: Union[str, Parametrized],
    ) -> None:
        """Changes the target qubit of a 'Local' channel.

        Args:
            qubits (Union[int, str, Iterable]): The new target for this
                channel. Must correspond to a qubit ID in device or an iterable
                of qubit IDs, when multi-qubit addressing is possible.
            channel (str): The channel's name provided when declared. Must be
                a channel with 'Local' addressing.
        """
        qubits = cast(QubitId, qubits)
        channel = cast(str, channel)

        self._target(qubits, channel)

    @_store
    def delay(
        self,
        duration: Union[int, Parametrized],
        channel: Union[str, Parametrized],
    ) -> None:
        """Idles a given channel for a specific duration.

        Args:
            duration (int): Time to delay (in multiples of 4 ns).
            channel (str): The channel's name provided when declared.
        """
        duration = cast(int, duration)
        channel = cast(str, channel)

        self._delay(duration, channel)

    @_store
    def measure(
        self, basis: Union[str, Parametrized] = "ground-rydberg"
    ) -> None:
        """Measures in a valid basis.

        Note:
            In addition to the supported bases of the selected device, allowed
            measurement bases will depend on the mode of operation. In
            particular, if using ``Microwave`` channels (XY mode), only
            measuring in the 'XY' basis is allowed. Inversely, it is not
            possible to measure in the 'XY' basis outside of XY mode.

        Args:
            basis (str): Valid basis for measurement (consult the
                ``supported_bases`` attribute of the selected device for
                the available options).
        """
        available = (
            self._device.supported_bases - {"XY"}
            if not self._in_xy
            else {"XY"}
        )
        if basis not in available:
            raise ValueError(
                f"The basis '{basis}' is not supported by the "
                "selected device and operation mode. The "
                "available options are: " + ", ".join(list(available))
            )

        if hasattr(self, "_measurement"):
            raise RuntimeError("The sequence has already been measured.")

        if self.is_parametrized():
            self._is_measured = True
        else:
            self._measurement = basis

    @_store
    def phase_shift(
        self,
        phi: Union[float, Parametrized],
        *targets: Union[QubitId, Parametrized],
        basis: Union[str, Parametrized] = "digital",
    ) -> None:
        r"""Shifts the phase of a qubit's reference by 'phi', for a given basis.

        This is equivalent to an :math:`R_z(\phi)` gate (i.e. a rotation of the
        target qubit's state by an angle :math:`\phi` around the z-axis of the
        Bloch sphere).

        Args:
            phi (float): The intended phase shift (in rads).
            targets (Union[int, str]): The ids of the qubits to apply the phase
                shift to.
            basis (str): The basis (i.e. electronic transition) to associate
                the phase shift to. Must correspond to the basis of a declared
                channel.
        """
        phi = cast(float, phi)
        basis = cast(str, basis)
        targets = cast(Tuple[QubitId], targets)

        self._phase_shift(phi, *targets, basis=basis)

    @_store
    def align(self, *channels: Union[str, Parametrized]) -> None:
        """Aligns multiple channels in time.

        Introduces delays that align the provided channels with the one that
        finished the latest, such that the next action added to any of them
        will start right after the latest channel has finished.

        Args:
            channels (str): The names of the channels to align, as given upon
                declaration.
        """
        ch_set = set(channels)
        # channels have to be a subset of the declared channels
        if not ch_set <= set(self._channels):
            raise ValueError(
                "All channel names must correspond to declared" " channels."
            )
        if len(channels) != len(ch_set):
            raise ValueError("The same channel was provided more than once.")

        if len(channels) < 2:
            raise ValueError("Needs at least two channels for alignment.")

        if self.is_parametrized():
            return

        last_ts = {id: self._last(cast(str, id)).tf for id in channels}
        tf = max(last_ts.values())

        for id in channels:
            delta = tf - last_ts[id]
            if delta > 0:
                self._delay(delta, cast(str, id))

    def build(self, **vars: Union[ArrayLike, float, int, str]) -> Sequence:
        """Builds a sequence from the programmed instructions.

        Keyword Args:
            vars: The values for all the variables declared in this Sequence
                instance, indexed by the name given upon declaration. Check
                ``Sequence.declared_variables`` to see all the variables.

        Returns:
            Sequence: The Sequence built with the given variable values.

        Example:
            ::

                # Check which variables are declared
                >>> print(seq.declared_variables)
                {'x': Variable(name='x', dtype=<class 'float'>, size=1),
                 'y': Variable(name='y', dtype=<class 'int'>, size=3)}
                # Build a sequence with specific values for both variables
                >>> seq1 = seq.build(x=0.5, y=[1, 2, 3])
        """
        if not self.is_parametrized():
            warnings.warn(
                "Building a non-parametrized sequence simply returns"
                " a copy of itself.",
                stacklevel=2,
            )
            return copy.copy(self)
        all_keys, given_keys = self._variables.keys(), vars.keys()
        if given_keys != all_keys:
            invalid_vars = given_keys - all_keys
            if invalid_vars:
                warnings.warn(
                    "No declared variables named: " + ", ".join(invalid_vars),
                    stacklevel=2,
                )
                for k in invalid_vars:
                    vars.pop(k, None)
            missing_vars = all_keys - given_keys
            if missing_vars:
                raise TypeError(
                    "Did not receive values for variables: "
                    + ", ".join(missing_vars)
                )

        for name, value in vars.items():
            self._variables[name]._assign(value)

        # Shallow copy with stored parametrized objects
        seq = copy.copy(self)
        # Eliminates the source of recursiveness errors
        seq._reset_parametrized()
        # Deepcopy the base sequence (what remains)
        seq = copy.deepcopy(seq)

        for call in self._to_build_calls:
            args_ = [
                arg.build() if isinstance(arg, Parametrized) else arg
                for arg in call.args
            ]
            kwargs_ = {
                key: val.build() if isinstance(val, Parametrized) else val
                for key, val in call.kwargs.items()
            }
            getattr(seq, call.name)(*args_, **kwargs_)

        return seq

    def serialize(self, **kwargs: Any) -> str:
        """Serializes the Sequence into a JSON formatted string.

        Other Parameters:
            kwargs: Valid keyword-arguments for ``json.dumps()``, except for
                ``cls``.

        Returns:
            str: The sequence encoded in a JSON formatted string.

        See Also:
            ``json.dumps``: Built-in function for serialization to a JSON
            formatted string.
        """
        return json.dumps(self, cls=PulserEncoder, **kwargs)

    @staticmethod
    def deserialize(obj: str, **kwargs: Any) -> Sequence:
        """Deserializes a JSON formatted string.

        Args:
            obj (str): The JSON formatted string to deserialize, coming from
                the serialization of a ``Sequence`` through
                ``Sequence.serialize()``.

        Other Parameters:
            kwargs: Valid keyword-arguments for ``json.loads()``, except for
                ``cls`` and ``object_hook``.

        Returns:
            Sequence: The deserialized Sequence object.

        See Also:
            ``json.loads``: Built-in function for deserialization from a JSON
            formatted string.
        """
        if "Sequence" not in obj:
            warnings.warn(
                "The given JSON formatted string does not encode a Sequence.",
                stacklevel=2,
            )

        return cast(Sequence, json.loads(obj, cls=PulserDecoder, **kwargs))

    @_screen
    def draw(
        self,
        draw_phase_area: bool = False,
        draw_interp_pts: bool = True,
        draw_phase_shifts: bool = False,
    ) -> None:
        """Draws the sequence in its current state.

        Keyword Args:
            draw_phase_area (bool): Whether phase and area values need to be
                shown as text on the plot, defaults to False.
            draw_interp_pts (bool): When the sequence has pulses with waveforms
                of type InterpolatedWaveform, draws the points of interpolation
                on top of the respective waveforms (defaults to True).
            draw_phase_shifts (bool): Whether phase shift and reference
                information should be added to the plot, defaults to False.

        See Also:
            Simulation.draw(): Draws the provided sequence and the one used by
            the solver.
        """
        draw_sequence(
            self,
            draw_phase_area=draw_phase_area,
            draw_interp_pts=draw_interp_pts,
            draw_phase_shifts=draw_phase_shifts,
        )

    def _target(
        self, qubits: Union[Iterable[QubitId], QubitId], channel: str
    ) -> None:
        self._validate_channel(channel)

        try:
            qubits_set = (
                set(cast(Iterable, qubits))
                if not isinstance(qubits, str)
                else {qubits}
            )
        except TypeError:
            qubits_set = {qubits}

        if self._channels[channel].addressing != "Local":
            raise ValueError("Can only choose target of 'Local' channels.")
        elif len(qubits_set) > cast(int, self._channels[channel].max_targets):
            raise ValueError(
                "This channel can target at most "
                f"{self._channels[channel].max_targets} qubits at a time"
            )

        if self.is_parametrized():
            for q in qubits_set:
                if q not in self._qids and not isinstance(q, Parametrized):
                    raise ValueError(
                        "All non-variable qubits must belong to the register."
                    )
            return

        elif not qubits_set.issubset(self._qids):
            raise ValueError("All given qubits must belong to the register.")

        basis = self._channels[channel].basis
        phase_refs = {self._phase_ref[basis][q].last_phase for q in qubits_set}
        if len(phase_refs) != 1:
            raise ValueError(
                "Cannot target multiple qubits with different "
                "phase references for the same basis."
            )

        try:
            last = self._last(channel)
            if last.targets == qubits_set:
                return
            ti = last.tf
            retarget = cast(int, self._channels[channel].retarget_time)
            elapsed = ti - self._last_target[channel]
            delta = cast(int, np.clip(retarget - elapsed, 0, retarget))
            if delta != 0:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    delta = self._channels[channel].validate_duration(
                        16 if delta < 16 else delta
                    )
            tf = ti + delta

        except ValueError:
            ti = -1
            tf = 0

        self._last_target[channel] = tf
        self._add_to_schedule(channel, _TimeSlot("target", ti, tf, qubits_set))

    def _delay(self, duration: int, channel: str) -> None:
        self._validate_channel(channel)
        if self.is_parametrized():
            return

        last = self._last(channel)
        ti = last.tf
        tf = ti + self._channels[channel].validate_duration(duration)
        self._add_to_schedule(
            channel, _TimeSlot("delay", ti, tf, last.targets)
        )

    def _phase_shift(
        self, phi: float, *targets: QubitId, basis: str = "digital"
    ) -> None:
        if basis not in self._phase_ref:
            raise ValueError("No declared channel targets the given 'basis'.")
        if self.is_parametrized():
            for t in targets:
                if t not in self._qids and not isinstance(t, Parametrized):
                    raise ValueError(
                        "All non-variable targets must belong to the register."
                    )
            return

        elif not set(targets) <= self._qids:
            raise ValueError(
                "All given targets have to be qubit ids declared"
                " in this sequence's register."
            )

        if phi % (2 * np.pi) == 0:
            return

        for qubit in targets:
            last_used = self._last_used[basis][qubit]
            new_phase = self._phase_ref[basis][qubit].last_phase + phi
            self._phase_ref[basis][qubit][last_used] = new_phase

    def _to_dict(self) -> dict[str, Any]:
        d = obj_to_dict(self, *self._calls[0].args, **self._calls[0].kwargs)
        d["__version__"] = pulser.__version__
        d["calls"] = self._calls[1:]
        d["vars"] = self._variables
        d["to_build_calls"] = self._to_build_calls
        return d

    def __str__(self) -> str:
        full = ""
        pulse_line = "t: {}->{} | {} | Targets: {}\n"
        target_line = "t: {}->{} | Target: {} | Phase Reference: {}\n"
        delay_line = "t: {}->{} | Delay \n"
        # phase_line = "t: {} | Phase shift of: {:.3f} | Targets: {}\n"
        for ch, seq in self._schedule.items():
            basis = self._channels[ch].basis
            full += f"Channel: {ch}\n"
            first_slot = True
            for ts in seq:
                if ts.type == "delay":
                    full += delay_line.format(ts.ti, ts.tf)
                    continue

                tgts = list(ts.targets)
                tgt_txt = ", ".join([str(t) for t in tgts])
                if isinstance(ts.type, Pulse):
                    full += pulse_line.format(ts.ti, ts.tf, ts.type, tgt_txt)
                elif ts.type == "target":
                    phase = self._phase_ref[basis][tgts[0]][ts.tf]
                    if first_slot:
                        full += (
                            f"t: 0 | Initial targets: {tgt_txt} | "
                            + f"Phase Reference: {phase} \n"
                        )
                        first_slot = False
                    else:
                        full += target_line.format(
                            ts.ti, ts.tf, tgt_txt, phase
                        )
            full += "\n"

        if hasattr(self, "_measurement"):
            full += f"Measured in basis: {self._measurement}"

        if self.is_parametrized():
            prelude = "Prelude\n-------\n" + full
            lines = ["Stored calls\n------------"]
            for i, c in enumerate(self._to_build_calls, 1):
                args = [str(a) for a in c.args]
                kwargs = [
                    f"{key}={str(value)}" for key, value in c.kwargs.items()
                ]
                lines.append(f"{i}. {c.name}({', '.join(args+kwargs)})")
            full = prelude + "\n\n".join(lines)

        return full

    def _add_to_schedule(self, channel: str, timeslot: _TimeSlot) -> None:
        if hasattr(self, "_measurement"):
            raise RuntimeError(
                "The sequence has already been measured. "
                "Nothing more can be added."
            )
        self._schedule[channel].append(timeslot)

    def _min_pulse_duration(self) -> float:
        duration_list = []
        for ch_schedule in self._schedule.values():
            for slot in ch_schedule:
                if isinstance(slot.type, Pulse):
                    duration_list.append(slot.tf - slot.ti)
        return min(duration_list)

    def _last(self, channel: str) -> _TimeSlot:
        """Shortcut to last element in the channel's schedule."""
        try:
            return self._schedule[channel][-1]
        except IndexError:
            raise ValueError("The chosen channel has no target.")

    def _validate_channel(self, channel: str) -> None:
        if channel not in self._channels:
            raise ValueError("Use the name of a declared channel.")

    def _validate_pulse(self, pulse: Pulse, channel: str) -> None:
        self._device.validate_pulse(pulse, self._taken_channels[channel])

    def _reset_parametrized(self) -> None:
        """Resets all attributes related to parametrization."""
        # Signals the sequence as actively "building" ie not parametrized
        self._building = True
        self._is_measured = False
        self._variables = {}
        self._to_build_calls = []


class _PhaseTracker:
    """Tracks a phase reference over time."""

    def __init__(self, initial_phase: float):
        self._times: list[int] = [0]
        self._phases: list[float] = [self._format(initial_phase)]

    @property
    def last_time(self) -> int:
        return self._times[-1]

    @property
    def last_phase(self) -> float:
        return self._phases[-1]

    def changes(
        self,
        ti: Union[float, int],
        tf: Union[float, int],
        time_scale: float = 1.0,
    ) -> Generator[tuple[float, float], None, None]:
        """Changes in phases within ]ti, tf]."""
        start, end = np.searchsorted(
            self._times, (ti * time_scale, tf * time_scale), side="right"
        )
        for i in range(start, end):
            change = self._phases[i] - self._phases[i - 1]
            yield (self._times[i] / time_scale, change)

    def _format(self, phi: float) -> float:
        return phi % (2 * np.pi)

    def __setitem__(self, t: int, phi: float) -> None:
        phase = self._format(phi)
        if t in self._times:
            ind = self._times.index(t)
            self._phases[ind] = phase
        else:
            ind = int(np.searchsorted(self._times, t, side="right"))
            self._times.insert(ind, t)
            self._phases.insert(ind, phase)

    def __getitem__(self, t: int) -> float:
        ind = int(np.searchsorted(self._times, t, side="right")) - 1
        return self._phases[ind]



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pulser.simulation import SimConfig


def test_init():
    config = SimConfig(
        noise=("SPAM", "doppler", "dephasing", "amplitude"),
        temperature=1000.0,
        runs=100,
    )
    str_config = config.__str__(True)
    assert (
        "SPAM, doppler, dephasing, amplitude" in str_config
        and "1000.0µK" in str_config
        and "100" in str_config
        and "Solver Options" in str_config
    )
    with pytest.raises(ValueError, match="is not a valid noise type."):
        SimConfig(noise="bad_noise")
    with pytest.raises(ValueError, match="Temperature field"):
        SimConfig(temperature=-1.0)
    with pytest.raises(ValueError, match="SPAM parameter"):
        SimConfig(eta=-1.0)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import numpy as np
import pytest

from pulser import Register
from pulser.devices import Chadoq2


def test_creation():
    empty_dict = {}
    with pytest.raises(ValueError, match="Cannot create a Register with"):
        Register(empty_dict)

    coords = [(0, 0), (1, 0)]
    ids = ["q0", "q1"]
    qubits = dict(zip(ids, coords))
    with pytest.raises(TypeError):
        Register(coords)
        Register(ids)

    with pytest.raises(ValueError, match="vectors of size 2 or 3"):
        Register.from_coordinates([(0, 1, 0, 1)])

    with pytest.raises(ValueError, match="vectors of size 2 or 3"):
        Register.from_coordinates([((1, 0),), ((-1, 0),)])

    reg1 = Register(qubits)
    reg2 = Register.from_coordinates(coords, center=False, prefix="q")
    assert np.all(np.array(reg1._coords) == np.array(reg2._coords))
    assert reg1._ids == reg2._ids

    reg3 = Register.from_coordinates(np.array(coords), prefix="foo")
    coords_ = np.array([(-0.5, 0), (0.5, 0)])
    assert reg3._ids == ["foo0", "foo1"]
    assert np.all(reg3._coords == coords_)
    assert not np.all(coords_ == coords)

    reg4 = Register.rectangle(1, 2, spacing=1)
    assert np.all(reg4._coords == coords_)

    reg5 = Register.square(2, spacing=2)
    coords_ = np.array([(-1, -1), (1, -1), (-1, 1), (1, 1)], dtype=float)
    assert np.all(np.array(reg5._coords) == coords_)

    reg6 = Register.triangular_lattice(2, 2, spacing=4)
    coords_ = np.array(
        [
            (-3, -np.sqrt(3)),
            (1, -np.sqrt(3)),
            (-1, np.sqrt(3)),
            (3, np.sqrt(3)),
        ]
    )
    assert np.all(np.array(reg6._coords) == coords_)


def test_rectangle():
    # Check rows
    with pytest.raises(ValueError, match="The number of rows"):
        Register.rectangle(0, 2)

    # Check columns
    with pytest.raises(ValueError, match="The number of columns"):
        Register.rectangle(2, 0)

    # Check spacing
    with pytest.raises(ValueError, match="Spacing"):
        Register.rectangle(2, 2, 0.0)


def test_square():
    # Check side
    with pytest.raises(ValueError, match="The number of atoms per side"):
        Register.square(0)

    # Check spacing
    with pytest.raises(ValueError, match="Spacing"):
        Register.square(2, 0.0)


def test_triangular_lattice():
    # Check rows
    with pytest.raises(ValueError, match="The number of rows"):
        Register.triangular_lattice(0, 2)

    # Check columns
    with pytest.raises(ValueError, match="The number of atoms per row"):
        Register.triangular_lattice(2, 0)

    # Check spacing
    with pytest.raises(ValueError, match="Spacing"):
        Register.triangular_lattice(2, 2, 0.0)


def test_hexagon():
    # Check number of layers
    with pytest.raises(ValueError, match="The number of layers"):
        Register.hexagon(0)

    # Check spacing
    with pytest.raises(ValueError, match="Spacing "):
        Register.hexagon(1, spacing=-1.0)

    # Check small hexagon (1 layer)
    reg = Register.hexagon(1, spacing=1.0)
    assert len(reg.qubits) == 7
    atoms = list(reg.qubits.values())
    crest_y = np.sqrt(3) / 2
    assert np.all(np.isclose(atoms[0], [0.0, 0.0]))
    assert np.all(np.isclose(atoms[1], [-0.5, crest_y]))
    assert np.all(np.isclose(atoms[2], [0.5, crest_y]))
    assert np.all(np.isclose(atoms[3], [1.0, 0.0]))
    assert np.all(np.isclose(atoms[4], [0.5, -crest_y]))
    assert np.all(np.isclose(atoms[5], [-0.5, -crest_y]))
    assert np.all(np.isclose(atoms[6], [-1.0, 0.0]))

    # Check a few atoms for a bigger hexagon (2 layers)
    reg = Register.hexagon(2, spacing=1.0)
    assert len(reg.qubits) == 19
    atoms = list(reg.qubits.values())
    crest_y = np.sqrt(3) / 2.0
    assert np.all(np.isclose(atoms[7], [-1.5, crest_y]))
    assert np.all(np.isclose(atoms[8], [-1.0, 2.0 * crest_y]))
    assert np.all(np.isclose(atoms[9], [-0.0, 2.0 * crest_y]))
    assert np.all(np.isclose(atoms[13], [1.5, -crest_y]))
    assert np.all(np.isclose(atoms[14], [1.0, -2.0 * crest_y]))
    assert np.all(np.isclose(atoms[15], [0.0, -2.0 * crest_y]))


def test_max_connectivity():
    device = Chadoq2
    max_atom_num = device.max_atom_num
    spacing = device.min_atom_distance
    crest_y = np.sqrt(3) / 2.0

    # Check device type
    with pytest.raises(TypeError):
        reg = Register.max_connectivity(2, None)

    # Check min number of atoms
    with pytest.raises(
        ValueError, match=r"The number of qubits(.+)greater than"
    ):
        reg = Register.max_connectivity(0, device)

    # Check max number of atoms
    with pytest.raises(ValueError, match=r"The number of qubits(.+)less than"):
        reg = Register.max_connectivity(max_atom_num + 1, device)

    # Check spacing
    reg = Register.max_connectivity(max_atom_num, device, spacing=spacing)
    with pytest.raises(ValueError, match="Spacing "):
        reg = Register.max_connectivity(
            max_atom_num, device, spacing=spacing - 1.0
        )

    # Check 1 atom
    reg = Register.max_connectivity(1, device)
    assert len(reg.qubits) == 1
    atoms = list(reg.qubits.values())
    assert np.all(np.isclose(atoms[0], [0.0, 0.0]))

    # Check for less than 7 atoms:
    for i in range(1, 7):
        hex_coords = np.array(
            [
                (0.0, 0.0),
                (-0.5, crest_y),
                (0.5, crest_y),
                (1.0, 0.0),
                (0.5, -crest_y),
                (-0.5, -crest_y),
            ]
        )
        reg = Register.max_connectivity(i, device)
        reg2 = Register.from_coordinates(
            spacing * hex_coords[:i], center=False
        )
        assert len(reg.qubits) == i
        atoms = list(reg.qubits.values())
        atoms2 = list(reg2.qubits.values())
        for k in range(i):
            assert np.all(np.isclose(atoms[k], atoms2[k]))

    # Check full layers on a small hexagon (1 layer)
    reg = Register.max_connectivity(7, device)
    assert len(reg.qubits) == 7
    atoms = list(reg.qubits.values())
    assert np.all(np.isclose(atoms[0], [0.0, 0.0]))
    assert np.all(np.isclose(atoms[1], [-0.5 * spacing, crest_y * spacing]))
    assert np.all(np.isclose(atoms[2], [0.5 * spacing, crest_y * spacing]))
    assert np.all(np.isclose(atoms[3], [1.0 * spacing, 0.0]))
    assert np.all(np.isclose(atoms[4], [0.5 * spacing, -crest_y * spacing]))
    assert np.all(np.isclose(atoms[5], [-0.5 * spacing, -crest_y * spacing]))
    assert np.all(np.isclose(atoms[6], [-1.0 * spacing, 0.0]))

    # Check full layers for a bigger hexagon (2 layers)
    reg = Register.max_connectivity(19, device)
    assert len(reg.qubits) == 19
    atoms = list(reg.qubits.values())
    assert np.all(np.isclose(atoms[7], [-1.5 * spacing, crest_y * spacing]))
    assert np.all(
        np.isclose(atoms[8], [-1.0 * spacing, 2.0 * crest_y * spacing])
    )
    assert np.all(np.isclose(atoms[13], [1.5 * spacing, -crest_y * spacing]))
    assert np.all(
        np.isclose(atoms[14], [1.0 * spacing, -2.0 * crest_y * spacing])
    )

    # Check extra atoms (2 full layers + 7 extra atoms)
    # for C3 symmetry, C6 symmetry and offset for next atoms
    reg = Register.max_connectivity(26, device)
    assert len(reg.qubits) == 26
    atoms = list(reg.qubits.values())
    assert np.all(np.isclose(atoms[19], [-2.5 * spacing, crest_y * spacing]))
    assert np.all(
        np.isclose(atoms[20], [-2.0 * spacing, 2.0 * crest_y * spacing])
    )
    assert np.all(
        np.isclose(atoms[21], [-0.5 * spacing, 3.0 * crest_y * spacing])
    )
    assert np.all(
        np.isclose(atoms[22], [2.0 * spacing, 2.0 * crest_y * spacing])
    )
    assert np.all(np.isclose(atoms[23], [2.5 * spacing, -crest_y * spacing]))
    assert np.all(
        np.isclose(atoms[24], [0.5 * spacing, -3.0 * crest_y * spacing])
    )
    assert np.all(
        np.isclose(atoms[25], [-2.0 * spacing, -2.0 * crest_y * spacing])
    )


def test_rotation():
    with pytest.raises(NotImplementedError):
        reg_ = Register.from_coordinates([(1, 0, 0), (0, 1, 4)])
        reg_.rotate(20)
    reg = Register.square(2, spacing=np.sqrt(2))
    reg.rotate(45)
    coords_ = np.array([(0, -1), (1, 0), (-1, 0), (0, 1)], dtype=float)
    assert np.all(np.isclose(reg._coords, coords_))


def test_drawing():
    with pytest.raises(NotImplementedError, match="register layouts in 2D."):
        reg_ = Register.from_coordinates([(1, 0, 0), (0, 1, 4)])
        reg_.draw()

    with pytest.raises(ValueError, match="Blockade radius"):
        reg = Register.from_coordinates([(1, 0), (0, 1)])
        reg.draw(blockade_radius=0.0)

    reg = Register.triangular_lattice(3, 8)
    with patch("matplotlib.pyplot.show"):
        reg.draw()

    reg = Register.rectangle(1, 8)
    with patch("matplotlib.pyplot.show"):
        reg.draw(blockade_radius=5, draw_half_radius=True, draw_graph=True)

    with pytest.raises(ValueError, match="'blockade_radius' to draw."):
        reg.draw(draw_half_radius=True)

    reg = Register.square(1)
    with pytest.raises(NotImplementedError, match="Needs more than one atom"):
        reg.draw(blockade_radius=5, draw_half_radius=True)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Decorators for adding parametrization support."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from itertools import chain
from typing import Any

from pulser.parametrized import Parametrized, ParamObj


def parametrize(func: Callable) -> Callable:
    """Makes a function support parametrized arguments.

    Note:
        Designed for use in class methods. Usage in instance or static methods
        is not supported, and in regular functions is not tested.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        for x in chain(args, kwargs.values()):
            if isinstance(x, Parametrized):
                return ParamObj(func, *args, **kwargs)
        return func(*args, **kwargs)

    return wrapper



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from unittest.mock import patch

import numpy as np
import pytest
from scipy.interpolate import interp1d, PchipInterpolator

from pulser.json.coders import PulserEncoder, PulserDecoder
from pulser.parametrized import Variable, ParamObj
from pulser.waveforms import (
    ConstantWaveform,
    KaiserWaveform,
    RampWaveform,
    BlackmanWaveform,
    CustomWaveform,
    CompositeWaveform,
    InterpolatedWaveform,
)

np.random.seed(20201105)

constant = ConstantWaveform(100, -3)
ramp = RampWaveform(2000, 5, 19)
arb_samples = np.random.random(52)
custom = CustomWaveform(arb_samples)
blackman = BlackmanWaveform(40, np.pi)
composite = CompositeWaveform(blackman, constant, custom)
interp_values = [0, 1, 4.4, 2, 3, 1, 0]
interp = InterpolatedWaveform(1000, interp_values)
kaiser = KaiserWaveform(40, np.pi)


def test_duration():
    with pytest.raises(TypeError, match="needs to be castable to an int"):
        ConstantWaveform("s", -1)
        RampWaveform([0, 1, 3], 1, 0)

    with pytest.raises(ValueError, match="positive duration"):
        ConstantWaveform(15, -10)
        RampWaveform(-20, 3, 4)

    with pytest.warns(UserWarning):
        wf = BlackmanWaveform(np.pi * 10, 1)

    assert wf.duration == 31
    assert custom.duration == 52
    assert composite.duration == 192


def test_change_duration():
    with pytest.raises(NotImplementedError):
        custom.change_duration(53)

    new_cte = constant.change_duration(103)
    assert constant.duration == 100
    assert new_cte.duration == 103

    new_blackman = blackman.change_duration(30)
    assert np.isclose(new_blackman.integral, blackman.integral)
    assert new_blackman != blackman

    new_ramp = ramp.change_duration(100)
    assert new_ramp.duration == 100
    assert new_ramp != ramp

    assert interp.duration == 1000
    new_interp = interp.change_duration(100)
    assert new_interp.duration == 100


def test_samples():
    assert np.all(constant.samples == -3)
    bm_samples = np.clip(np.blackman(40), 0, np.inf)
    bm_samples *= np.pi / np.sum(bm_samples) / 1e-3
    comp_samples = np.concatenate([bm_samples, np.full(100, -3), arb_samples])
    assert np.all(np.isclose(composite.samples, comp_samples))


def test_integral():
    assert np.isclose(blackman.integral, np.pi)
    assert constant.integral == -0.3
    assert ramp.integral == 24


def test_draw():
    with patch("matplotlib.pyplot.show"):
        composite.draw()
        blackman.draw()
        interp.draw()


def test_eq():
    assert constant == CustomWaveform(np.full(100, -3))
    assert constant != -3
    assert constant != CustomWaveform(np.full(48, -3))


def test_first_last():
    assert constant.first_value == constant.last_value
    assert ramp.first_value == 5
    assert ramp.last_value == 19
    assert blackman.first_value == 0
    assert blackman.last_value == 0
    assert composite.first_value == 0
    assert composite.last_value == arb_samples[-1]
    assert custom.first_value == arb_samples[0]
    assert np.isclose(interp.first_value, interp_values[0])
    assert np.isclose(interp.last_value, interp_values[-1])


def test_hash():
    assert hash(constant) == hash(tuple(np.full(100, -3)))
    assert hash(ramp) == hash(tuple(np.linspace(5, 19, num=2000)))


def test_composite():
    with pytest.raises(ValueError, match="Needs at least two waveforms"):
        CompositeWaveform()
        CompositeWaveform(composite)
        CompositeWaveform([blackman, custom])
        CompositeWaveform(10)

    with pytest.raises(TypeError, match="not a valid waveform"):
        CompositeWaveform(composite, "constant")

    assert composite.waveforms == [blackman, constant, custom]

    wf = CompositeWaveform(blackman, constant)
    msg = (
        "BlackmanWaveform(40 ns, Area: 3.14), "
        + "ConstantWaveform(100 ns, -3 rad/µs)"
    )
    assert wf.__str__() == f"Composite({msg})"
    assert wf.__repr__() == f"CompositeWaveform(140 ns, [{msg}])"


def test_custom():
    data = np.arange(16, dtype=float)
    wf = CustomWaveform(data)
    assert wf.__str__() == "Custom"
    assert wf.__repr__() == f"CustomWaveform(16 ns, {data!r})"


def test_ramp():
    assert ramp.slope == 7e-3


def test_blackman():
    with pytest.raises(TypeError):
        BlackmanWaveform(100, np.array([1, 2]))
    wf = BlackmanWaveform(100, -2)
    assert np.isclose(wf.integral, -2)
    assert np.all(wf.samples <= 0)
    assert wf == BlackmanWaveform(100, np.array([-2]))

    with pytest.raises(ValueError, match="matching signs"):
        BlackmanWaveform.from_max_val(-10, np.pi)

    wf = BlackmanWaveform.from_max_val(10, 2 * np.pi)
    assert np.isclose(wf.integral, 2 * np.pi)
    assert np.max(wf.samples) < 10

    wf = BlackmanWaveform.from_max_val(-10, -np.pi)
    assert np.isclose(wf.integral, -np.pi)
    assert np.min(wf.samples) > -10

    var = Variable("var", float)
    wf_var = BlackmanWaveform.from_max_val(-10, var)
    assert isinstance(wf_var, ParamObj)
    var._assign(-np.pi)
    assert wf_var.build() == wf

    # Check moving back to previous even duration
    area: float = np.pi / 6
    max_val: float = 46
    wf: BlackmanWaveform = BlackmanWaveform.from_max_val(max_val, area)
    duration = wf.duration
    assert duration % 2 == 0
    wf2 = BlackmanWaveform(duration + 1, area)
    assert np.max(wf2.samples) < np.max(wf.samples) <= max_val

    # Same but with a negative max value
    wf: BlackmanWaveform = BlackmanWaveform.from_max_val(-max_val, -area)
    duration = wf.duration
    assert duration % 2 == 0
    wf2 = BlackmanWaveform(duration + 1, -area)
    assert np.min(wf2.samples) > np.min(wf.samples) >= -max_val


def test_interpolated():
    assert isinstance(interp.interp_function, PchipInterpolator)

    times = np.linspace(0.2, 0.8, num=len(interp_values))
    with pytest.raises(ValueError, match="must match the number of `values`"):
        InterpolatedWaveform(1000, interp_values, times=times[:-1])
    with pytest.raises(ValueError, match="must be greater than or equal to 0"):
        InterpolatedWaveform(1000, interp_values, times=times - 0.21)
    with pytest.raises(ValueError, match="must be less than or equal to 1"):
        InterpolatedWaveform(1000, interp_values, times=times + 0.21)
    with pytest.raises(ValueError, match="array of non-repeating values"):
        InterpolatedWaveform(
            1000, interp_values, times=[0.2] + times[:-1].tolist()
        )

    with pytest.raises(ValueError, match="Invalid interpolator 'fake'"):
        InterpolatedWaveform(
            1000, interp_values, times=times, interpolator="fake"
        )

    dt = 1000
    interp_wf = InterpolatedWaveform(
        dt, [0, 1], interpolator="interp1d", kind="linear"
    )
    assert isinstance(interp_wf.interp_function, interp1d)
    np.testing.assert_allclose(interp_wf.samples, np.linspace(0, 1.0, num=dt))

    interp_wf *= 2
    np.testing.assert_allclose(interp_wf.samples, np.linspace(0, 2.0, num=dt))

    wf_str = "InterpolatedWaveform(Points: (0, 0), (999, 2)"
    assert str(interp_wf) == wf_str + ")"
    assert repr(interp_wf) == wf_str + ", Interpolator=interp1d)"

    vals = np.linspace(0, 1, num=5) ** 2
    interp_wf2 = InterpolatedWaveform(
        dt, vals, interpolator="interp1d", kind="quadratic"
    )
    np.testing.assert_allclose(
        interp_wf2.samples, np.linspace(0, 1, num=dt) ** 2, atol=1e-3
    )


def test_kaiser():
    duration: int = 40
    area: float = np.pi
    beta: float = 14.0

    wf: KaiserWaveform = KaiserWaveform(duration, area, beta)

    # Check type error on area
    with pytest.raises(TypeError):
        KaiserWaveform(duration, np.array([1, 2]))

    # Check type error on beta
    with pytest.raises(TypeError):
        KaiserWaveform(duration, area, beta=np.array([1, 2]))

    # Check beta must not be negative
    with pytest.raises(ValueError, match="must be greater than 0"):
        KaiserWaveform(duration, area, -1.0)

    # Check duration
    assert wf.duration == duration
    assert wf.samples.size == duration

    # Check default beta
    wf_default_beta = KaiserWaveform(duration, area)
    kaiser_beta_14: np.ndarray = np.kaiser(duration, 14.0)
    kaiser_beta_14 *= area / float(np.sum(kaiser_beta_14)) / 1e-3
    np.testing.assert_allclose(
        wf_default_beta.samples, kaiser_beta_14, atol=1e-3
    )

    # Check area
    assert np.isclose(np.sum(wf.samples), area * 1000.0)

    # Check duration change
    new_duration = duration * 2
    wf_change_duration = wf.change_duration(new_duration)
    assert wf_change_duration.samples.size == new_duration
    assert np.isclose(np.sum(wf.samples), np.sum(wf_change_duration.samples))

    # Check __str__
    assert str(wf) == (
        f"Kaiser({duration} ns, Area: {area:.3g}, Beta: {beta:.3g})"
    )

    # Check __repr__
    assert repr(wf) == (
        f"KaiserWaveform(duration: {duration}, "
        f"area: {area:.3g}, beta: {beta:.3g})"
    )

    # Check multiplication
    wf_multiplication = wf * 2
    assert (wf_multiplication.samples == wf.samples * 2).all()

    # Check area and max_val must have matching signs
    with pytest.raises(ValueError, match="must have matching signs"):
        KaiserWaveform.from_max_val(1, -1)

    # Test from_max_val
    for max_val in range(1, 501, 50):
        for beta in range(1, 20):
            wf = KaiserWaveform.from_max_val(max_val, area, beta)
            assert np.isclose(np.sum(wf.samples), area * 1000.0)
            assert np.max(wf.samples) <= max_val
            wf = KaiserWaveform.from_max_val(-max_val, -area, beta)
            assert np.isclose(np.sum(wf.samples), -area * 1000.0)
            assert np.min(wf.samples) >= -max_val


def test_ops():
    assert -constant == ConstantWaveform(100, 3)
    assert ramp * 2 == RampWaveform(2e3, 10, 38)
    assert --custom == custom
    assert blackman / 2 == BlackmanWaveform(40, np.pi / 2)
    assert composite * 1 == composite
    with pytest.raises(ZeroDivisionError):
        constant / 0


def test_serialization():
    for wf in [blackman, composite, constant, custom, interp, kaiser, ramp]:
        s = json.dumps(wf, cls=PulserEncoder)
        assert wf == json.loads(s, cls=PulserDecoder)


def test_get_item():

    # Check errors raised

    duration = constant.duration
    with pytest.raises(
        IndexError,
        match=re.escape(
            "Index ('index_or_slice' = "
            f"{duration}) must be in the range "
            f"0~{duration-1}, or "
            f"{-duration}~-1 from the end."
        ),
    ):
        constant[duration]
    with pytest.raises(
        IndexError,
        match=re.escape(
            "Index ('index_or_slice' = "
            f"{-duration-1}) must be in the range "
            f"0~{duration-1}, or "
            f"{-duration}~-1 from the end."
        ),
    ):
        constant[-duration - 1]

    with pytest.raises(
        IndexError, match="The step of the slice must be None or 1."
    ):
        constant[0:1:2]

    # Check nominal operations

    for wf in [blackman, composite, constant, custom, kaiser, ramp, interp]:
        duration = wf.duration
        duration14 = duration // 4
        duration34 = duration * 3 // 4
        samples = wf.samples

        # Check with int index
        for i in range(-duration, duration):
            assert wf[i] == samples[i]

        # Check with slices

        assert (wf[0:duration] == samples).all()
        assert (wf[0:-1] == samples[0:-1]).all()
        assert (wf[0:] == samples).all()
        assert (wf[-1:] == samples[-1:]).all()
        assert (wf[:duration] == samples).all()
        assert (wf[:] == samples).all()
        assert (
            wf[duration14:duration34] == samples[duration14:duration34]
        ).all()
        assert (
            wf[-duration34:-duration14] == samples[-duration34:-duration14]
        ).all()

        # Check with out of bounds slices
        assert (wf[: duration * 2] == samples).all()
        assert (wf[-duration * 2 :] == samples).all()
        assert (wf[-duration * 2 : duration * 2] == samples).all()
        assert (
            wf[duration // 2 : duration * 2]
            == samples[duration // 2 : duration * 2]
        ).all()
        assert (
            wf[-duration * 2 : duration // 2]
            == samples[-duration * 2 : duration // 2]
        ).all()
        assert wf[2:1].size == 0
        assert wf[duration * 2 :].size == 0
        assert wf[duration * 2 : duration * 3].size == 0
        assert wf[-duration * 3 : -duration * 2].size == 0



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import numpy as np
import pytest

from pulser import Sequence, Register, Pulse
from pulser.devices import Chadoq2, MockDevice
from pulser.parametrized import Variable
from pulser.waveforms import BlackmanWaveform

reg = Register.rectangle(4, 3)
device = Chadoq2


def test_var_declarations():
    sb = Sequence(reg, device)
    assert sb.declared_variables == {}
    var = sb.declare_variable("var")
    assert sb.declared_variables == {"var": var}
    assert isinstance(var, Variable)
    assert var.dtype == float
    assert len(var) == 1
    with pytest.raises(ValueError, match="already being used"):
        sb.declare_variable("var", dtype=int, size=10)
    var2 = sb.declare_variable("var2", 4, str)
    assert var2.dtype == str
    assert len(var2) == 4


def test_stored_calls():
    sb = Sequence(reg, device)
    assert sb._calls[-1].name == "__init__"
    var = sb.declare_variable("var")
    assert sb._to_build_calls == []
    sb.declare_channel("ch1", "rydberg_local", initial_target=var)
    assert sb._calls[-1].name == "declare_channel"
    assert sb._to_build_calls[-1].name == "target"
    assert sb._to_build_calls[-1].args == (var, "ch1")
    with pytest.raises(ValueError, match="name of a declared channel"):
        sb.delay(1000, "rydberg_local")
    x = Variable("x", str)
    var_ = copy.deepcopy(var)
    with pytest.raises(ValueError, match="Unknown variable 'x'"):
        sb.target(x, "ch1")
    with pytest.raises(ValueError, match="come from this Sequence"):
        sb.target(var_, "ch1")

    with pytest.raises(ValueError, match="non-variable qubits must belong"):
        sb.target("q20", "ch1")

    sb.delay(var, "ch1")
    call = sb._to_build_calls[1]
    assert call.name == "delay"
    assert call.args == (var, "ch1")
    assert call.kwargs == {}

    pls = Pulse.ConstantPulse(1000, var, var, var)

    with pytest.raises(ValueError, match="Invalid protocol 'last'"):
        sb.add(pls, "ch1", protocol="last")

    with pytest.raises(ValueError, match="amplitude goes over the maximum"):
        sb.add(
            Pulse.ConstantPulse(20, 2 * np.pi * 100, -2 * np.pi * 100, 0),
            "ch1",
        )
    with pytest.raises(
        ValueError, match="detuning values go out of the range"
    ):
        sb.add(Pulse.ConstantPulse(500, 2 * np.pi, -2 * np.pi * 100, 0), "ch1")

    assert sb._to_build_calls[-1] == call
    sb.add(pls, "ch1", protocol="wait-for-all")
    call = sb._to_build_calls[2]
    assert call.name == "add"
    assert call.args == (pls, "ch1")
    assert call.kwargs == {"protocol": "wait-for-all"}

    q_var = sb.declare_variable("q_var", size=5, dtype=str)
    sb.declare_channel("ch2", "rydberg_global")
    assert len(sb._calls) == 3
    assert sb._calls[-1].name == "declare_channel"
    with pytest.raises(ValueError, match="'Local' channels"):
        sb.target(0, "ch2")
    with pytest.raises(ValueError, match="target at most 1 qubits"):
        sb.target(q_var, "ch1")

    sb2 = Sequence(reg, MockDevice)
    sb2.declare_channel("ch1", "rydberg_local", initial_target={3, 4, 5})
    q_var2 = sb2.declare_variable("q_var2", size=5, dtype=str)
    var2 = sb2.declare_variable("var2")
    assert sb2._building
    sb2.target({var2, 7, 9, 10}, "ch1")
    assert not sb2._building
    sb2.target(q_var2, "ch1")

    with pytest.raises(ValueError, match="targets the given 'basis'"):
        sb.phase_shift(var, *q_var)

    with pytest.raises(ValueError, match="non-variable targets must belong"):
        sb.phase_shift(var, *q_var, "q1", basis="ground-rydberg")

    with pytest.raises(ValueError, match="correspond to declared channels"):
        sb.align("ch1", var)
    with pytest.raises(ValueError, match="more than once"):
        sb.align("ch1", "ch2", "ch2")
    with pytest.raises(ValueError, match="at least two channels"):
        sb.align("ch1")

    with pytest.raises(ValueError, match="not supported"):
        sb.measure(basis="z")
    sb.measure()
    with pytest.raises(RuntimeError):
        sb.delay(var * 50, "ch1")


def test_build():
    reg_ = Register.rectangle(2, 1, prefix="q")
    sb = Sequence(reg_, device)
    var = sb.declare_variable("var")
    targ_var = sb.declare_variable("targ_var", size=2, dtype=str)
    sb.declare_channel("ch1", "rydberg_local")
    sb.declare_channel("ch2", "raman_local", initial_target=targ_var[0])
    sb.target(targ_var[1], "ch1")
    wf = BlackmanWaveform(var * 100, np.pi)
    pls = Pulse.ConstantDetuning(wf, var, var)
    sb.add(pls, "ch1")
    sb.delay(var * 50, "ch1")
    sb.align("ch2", "ch1")
    sb.phase_shift(var, targ_var[0])
    pls2 = Pulse.ConstantPulse(wf.duration, var, var, 0)
    sb.add(pls2, "ch2")
    sb.measure()
    with pytest.warns(UserWarning, match="No declared variables"):
        sb.build(t=100, var=2, targ_var=["q1", "q0"])
    with pytest.raises(TypeError, match="Did not receive values for"):
        sb.build(var=2)
    seq = sb.build(var=2, targ_var=["q1", "q0"])
    assert seq._schedule["ch2"][-1].tf == 500
    assert seq.current_phase_ref("q1") == 2.0
    assert seq.current_phase_ref("q0") == 0.0
    assert seq._measurement == "ground-rydberg"

    s = sb.serialize()
    sb_ = Sequence.deserialize(s)
    assert str(sb) == str(sb_)

    s2 = sb_.serialize()
    sb_2 = Sequence.deserialize(s2)
    assert str(sb) == str(sb_2)


def test_str():
    reg_ = Register.rectangle(2, 1, prefix="q")
    sb = Sequence(reg_, device)
    sb.declare_channel("ch1", "rydberg_global")
    with pytest.warns(UserWarning, match="Building a non-parametrized"):
        seq = sb.build()
    var = sb.declare_variable("var")
    pls = Pulse.ConstantPulse(var * 100, var, -1, var)
    sb.add(pls, "ch1")
    s = (
        f"Prelude\n-------\n{str(seq)}Stored calls\n------------\n\n"
        + "1. add(Pulse.ConstantPulse(mul(var, 100), var, -1, var), ch1)"
    )
    assert s == str(sb)


def test_screen():
    sb = Sequence(reg, device)
    sb.declare_channel("ch1", "rydberg_global")
    assert sb.current_phase_ref(4, basis="ground-rydberg") == 0
    var = sb.declare_variable("var")
    sb.delay(var, "ch1")
    with pytest.raises(RuntimeError, match="can't be called in parametrized"):
        sb.current_phase_ref(4, basis="ground-rydberg")
